// TGoalTrack.cpp: implementation of the TGoalTrack class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "NetworkHeader.h"
#include "TScreenParamsNetwork.h"
#include "TGoalTrack.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

TGoalTrack::TGoalTrack(int iNoOfGoals)
{
	int	i;

	m_iNoOfGoals = iNoOfGoals;
	m_dSumActivation = 0.0;
	m_dAbsActivation = 0.0;
	m_pdPercent = new double[iNoOfGoals];
	m_pdActivation = new double[iNoOfGoals];
	
	for (i = 0; i < iNoOfGoals; i++)
	{
		m_pdActivation[i] = 0.0;
		m_pdPercent[i] = 0.0;
	}
}

TGoalTrack::~TGoalTrack()
{
	delete [] m_pdActivation;
	delete [] m_pdPercent;
}

/*-----------------------------------------------------------------------------
Function: void TGoalTrack::ReduceActivation(double beta)
Date    : 07.07.98
-------------------------------------------------------------------------------
Parameter:	activation reduction beta
Returnval: (void)
Description: called to decrease the sum of activation by beta as well as to 
				 take into account the influence of internal spreading to the percent
				 of goal influence. This is done after activation spreading to be
				 independent of the order in which modules do activation spreading
-----------------------------------------------------------------------------*/
void TGoalTrack::ReduceActivation(double beta)
{
	int		i;
	double	value;

	// take previously spreading activation into account
	if (m_dAbsActivation > D_ZERO_P)
	{
		for (i = 0; i < m_iNoOfGoals; i++)
		{
			// update the percentage, which is always positive
			value = m_pdActivation[i];
//			if (value < 0.0)
//				value = -value;
			// add previous influence
			value += m_pdPercent[i] * m_dSumActivation;
			m_pdPercent[i] = value / m_dAbsActivation;
			m_pdActivation[i] = 0.0;
		}
	}
	else
		for (i = 0; i < m_iNoOfGoals; i++)
			m_pdActivation[i] = 0.0;

	m_dSumActivation = m_dAbsActivation * beta;
}

/*-----------------------------------------------------------------------------
Function: void TGoalTrack::SetActivation(int i, double val)
Date    : 08.07.98
-------------------------------------------------------------------------------
Parameter: index of the goal
			  value for the activation
Returnval: (void)
Description: sets the new activation, which has been received direct from a goal
-----------------------------------------------------------------------------*/
void TGoalTrack::SetActivation(int i, double val)
{
	ASSERT ((i >= 0) && (i < m_iNoOfGoals)); 
	m_pdActivation[i] = val;
}

/*-----------------------------------------------------------------------------
Function: void TGoalTrack::SetGoalTrack(TGoalTrack* pSource)
Date    : 08.07.98
-------------------------------------------------------------------------------
Parameter: pointer to the source
Returnval: (void)
Description: copies the contents of pSource
-----------------------------------------------------------------------------*/
void TGoalTrack::SetGoalTrack(TGoalTrack* pSource)
{
	int	i;

	ASSERT(m_iNoOfGoals == pSource->GetNoOfGoals());
	m_dSumActivation = pSource->GetSumActivation();
	for (i = 0; i < m_iNoOfGoals; i++)
	{
		m_pdPercent[i] = pSource->GetPercent(i);
		m_pdActivation[i] = pSource->GetActivation(i);
	}
}

/*-----------------------------------------------------------------------------
Function: void TGoalTrack::SetAbsActivationFromPercentage(double dActivation)
Date    : 08.07.98
-------------------------------------------------------------------------------
Parameter: the activation which is 100 percent
Returnval: (void)
Description: sets the percentage values of activation to absolute activations
-----------------------------------------------------------------------------*/
double TGoalTrack::SetAbsActivationFromPercentage(double dActivation)
{
	int		i;
	double	newActivation = dActivation;


	for (i = 0; i < m_iNoOfGoals; i++)
	{
		m_pdActivation[i] = m_pdPercent[i] * dActivation;
		newActivation -= m_pdActivation[i];
	}

	// keep track of the absolute activation
	if (newActivation < 0.0)
		m_dAbsActivation -= newActivation;
	else
		m_dAbsActivation += newActivation;

	return newActivation;
}

/*-----------------------------------------------------------------------------
Function: void TGoalTrack::UpdateActivation(TGoalTrack* pSource)
Date    : 08.07.98
-------------------------------------------------------------------------------
Parameter: pointer to the other goal track
Returnval: (void)
Description: values of this goal activations are set to the source tracks values,
				 if they are greater
-----------------------------------------------------------------------------*/
void TGoalTrack::UpdateActivation(TGoalTrack* pSource)
{
	int		i;
	double	dActivation;
	
	for (i = 0; i < m_iNoOfGoals; i++)
	{
		// function for updating is max(abs), the maximum absolute value is stored
		dActivation = pSource->GetActivation(i);
		if (dActivation < 0.0)
			dActivation = -dActivation;

		 if (m_pdActivation[i] < dActivation)
			// store the signed value
			m_pdActivation[i] = pSource->GetActivation(i);
	}
}

/*-----------------------------------------------------------------------------
Function: double TGoalTrack::GetTrackedActivation()
Date    : 08.07.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: returns the sum of all activations that are the most influencial
				 activations from goals
-----------------------------------------------------------------------------*/
double TGoalTrack::GetTrackedActivation()
{
	int		i;
	double	sum = 0.0;
	double	activation;

	for (i = 0; i < m_iNoOfGoals; i++)
	{
		// tracked activation
		activation = m_pdActivation[i];
		sum += activation;

		// keep track as well of the absolute influence
		if (activation < 0.0) 
			m_dAbsActivation -= activation;
		else
			m_dAbsActivation += activation;
	}
	
	return sum;
}

/*-----------------------------------------------------------------------------
Function: void TGoalTrack::Draw(CDC* pDC, TScreenParamsNetwork* pScreen, CPoint& coords)
Date    : 08.07.98
-------------------------------------------------------------------------------
Parameter: device context
			  screnn parameters
			  coordinates of the center of the module
Returnval: (void)
Description: draws the influcence of each goal to this module
-----------------------------------------------------------------------------*/
void TGoalTrack::Draw(CDC* pDC, TScreenParamsNetwork* pScreen, CPoint& coords)
{
	int		i;
	int		nPointSize;
	CFont		font;
	CFont*	oldFont;
	CRect		field;
	COLORREF	oldTextColor;
	COLORREF	oldBkColor;
	int		sizeX;
	CString	strActivation;
	int		x = coords.x;
	int		y = coords.y + 3;

	// create the font
	nPointSize = (int) (30.0 * pScreen->GetScale());
	if (!font.CreatePointFont(nPointSize, _T("Times New Roman"), pDC))
		font.CreateStockObject(SYSTEM_FONT);
	oldFont = pDC->SelectObject(&font);

	// calculate the rectangle of the text
	oldBkColor = pDC->SetBkColor(0x00ffffff);
	oldTextColor = pDC->SetTextColor(0x00000000);
	sizeX = (int) pScreen->GetScale() * 5 * nPointSize / 20;

	for (i = 0; i < m_iNoOfGoals; i++)
	{
		// write each goal
		y += nPointSize / 7 + 2;
		field.SetRect(x, y, x + sizeX, y + nPointSize / 7);
		strActivation.Format("%4.3f", m_pdPercent[i]);
		pDC->DrawText(strActivation, field, DT_SINGLELINE | DT_VCENTER);
	}

	pDC->SetTextColor(oldTextColor);
	pDC->SetBkColor(oldBkColor);

	pDC->SelectObject(oldFont);
	font.DeleteObject();
}