// TPercept.cpp: implementation of the TPercept class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "NetworkHeader.h"
#include "TProposition.h"
#include "TWorldState.h"
#include "TPercept.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

TPercept::TPercept()
{
	m_pFactoryObjects = new CObList();
}


TPercept::~TPercept()
{
	while (!m_pFactoryObjects->IsEmpty())
		delete m_pFactoryObjects->RemoveHead();
	delete m_pFactoryObjects;
}


/*-----------------------------------------------------------------------------
Function: void TPercept::InitializeAll()
Date    : 22.01.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: initializes all world states (e.g. as being not used)
-----------------------------------------------------------------------------*/
void TPercept::InitializeAll()
{
	POSITION pos = m_pFactoryObjects->GetHeadPosition();

	while (pos != NULL)
		((TWorldState*) m_pFactoryObjects->GetNext(pos))->Initialize();
}

/*-----------------------------------------------------------------------------
Function: int TPercept::AddPerception(LPCTSTR name, CObject* pObj, PCallBack pCallback)
Date    : 22.10.98
-------------------------------------------------------------------------------
Parameter: name of the Proposition
			  pointer to the object that contains the truth value function
			  pointer to the truth value function to call
Returnval: index of the added Proposition
Description: creates and adds a new TWorld state for this proposition
-----------------------------------------------------------------------------*/
int TPercept::AddPerception(LPCTSTR name, CObject* pObj, PCallBackPerception pCallback)
{
	TWorldState* pState;

	pState = new TWorldState(name, m_pFactoryObjects->GetCount(), pObj, pCallback);
	m_pFactoryObjects->AddTail(pState);
	return m_pFactoryObjects->GetCount();
}

/*-----------------------------------------------------------------------------
Function: void TPercept::CalculateAllActivations()
Date    : 22.01.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: calculates the activation of all percepts
-----------------------------------------------------------------------------*/
void TPercept::CalculateAllActivations()
{
	POSITION			pos = m_pFactoryObjects->GetHeadPosition();
	TWorldState*	pState;
	
	while (pos != NULL)
	{
		pState = (TWorldState*) m_pFactoryObjects->GetNext(pos);
		if (pState->IsUsed())
			pState->GetTruthValue();
	}
}

/*-----------------------------------------------------------------------------
Function: void TPercept::WriteActivationDistribution(CArchive& ar)
Date    : 16.02.99
-------------------------------------------------------------------------------
Parameter: archive to write to
Returnval: (void)
Description: writes the activation distribution of all used perceptions
-----------------------------------------------------------------------------*/
void TPercept::WriteActivationDistribution(CArchive& ar)
{
	POSITION			pos = m_pFactoryObjects->GetHeadPosition();
	TWorldState*	pState;
	CString			initialLine;
	
	initialLine = "Name .1 .2 .3 .4 .5 .6 .7 .8 .9 1\r\n";
	ar.WriteString(initialLine);
	while (pos != NULL)
	{
		pState = (TWorldState*) m_pFactoryObjects->GetNext(pos);
		if (pState->IsUsed())
			pState->WriteActivationDistribution(ar);
	}
}

/*-----------------------------------------------------------------------------
Function: void TPercept::DrawAll(CDC* pDC, TScreenParamsNetwork* pScreen)
Date    : 02.02.98
-------------------------------------------------------------------------------
Parameter: pointer to the device context
			  pointer to the screen parameters as translation, scaling
Returnval: (void)
Description: draws each perception module
-----------------------------------------------------------------------------*/
void TPercept::DrawAll(CDC* pDC, TScreenParamsNetwork* pScreen)
{
	POSITION			pos = m_pFactoryObjects->GetHeadPosition();
	TWorldState*	pState;
	
	while (pos != NULL)
	{
		pState = (TWorldState*) m_pFactoryObjects->GetNext(pos);
		if (pState->IsUsed())
			pState->DrawModule(pDC, pScreen);
	}
}

/*-----------------------------------------------------------------------------
Function: void TPercept::SetAllCoordinates(TScreenParamsNetwork* pScreen)
Date    : 04.02.98
-------------------------------------------------------------------------------
Parameter: pointer to the screen parameters as translation, scaling
Returnval: (void)
Description: sets each perception modules coordinates
-----------------------------------------------------------------------------*/
void TPercept::SetAllCoordinates(TScreenParamsNetwork* pScreen)
{
	POSITION			pos = m_pFactoryObjects->GetHeadPosition();
	TWorldState*	pState;
	
	while (pos != NULL)
	{
		pState = (TWorldState*) m_pFactoryObjects->GetNext(pos);
		if (pState->IsUsed())
			pState->SetCoordinates(pScreen);
	}
}

/*-----------------------------------------------------------------------------
Function: void TPercept::SetAllNormCoordinates(CPoint coord, CSize size, int iMaxX)
Date    : 02.02.98
-------------------------------------------------------------------------------
Parameter: scaling of x coordinate positions (default 4)
			  translation of y coordinate positions (default 0)
			  x coordinate of rightmost competence module
Returnval: (void)
Description: sets the norm coordinates of all perceptions that are used by the 
				 current network
-----------------------------------------------------------------------------*/
void TPercept::SetAllNormCoordinates(CPoint coord, CSize size, int iMaxX)
{
	POSITION			pos = m_pFactoryObjects->GetHeadPosition();
	TWorldState*	pState;
	int				noOfUsedPerceptions = 0;
	int				noOfColumns = 0;
	int				x;
	
	while (pos != NULL)
	{
		pState = (TWorldState*) m_pFactoryObjects->GetNext(pos);
	
		if (pState->IsUsed())
		{
			x = (noOfUsedPerceptions + 1) * size.cx * (int) MODULE_SIZE + coord.x;
			if (x > iMaxX)
			{
				// start a new column
				noOfColumns++;
				noOfUsedPerceptions = 0;
				x = (noOfUsedPerceptions + 1) * size.cx * (int) MODULE_SIZE;
			}

			pState->SetNormCoordinates(CPoint(x, coord.y + size.cy + noOfColumns * 15));
			noOfUsedPerceptions++;
		}
	}
}


/*-----------------------------------------------------------------------------
Function: int TPercept::GetPerceptionIndex(LPCTSTR name)
Date    : 14.01.98
-------------------------------------------------------------------------------
Parameter: name of the perception
Returnval: pointer to the Worldstate (Perception) NULL, if name does not exist
Description: searches for the perception in the list of perception names.
-----------------------------------------------------------------------------*/
TWorldState* TPercept::GetPerception(LPCTSTR name)
{	
	POSITION			pos = m_pFactoryObjects->GetHeadPosition();
	TWorldState*	pState;

	while (pos != NULL)
	{
		pState = (TWorldState*) m_pFactoryObjects->GetNext(pos);
		if((pState->GetName().CompareNoCase(name)) == 0)
			return pState;
	}
	return NULL;
}

/*-----------------------------------------------------------------------------
Function: void TPercept::SetUseFuzzyValues(BOOL val)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: flag indicating if fuzzy or bool values are used for worldstates
Returnval: (void)
Description: sets all worldstates
-----------------------------------------------------------------------------*/
void TPercept::SetUseFuzzyValues(BOOL val)
{
	POSITION	pos = m_pFactoryObjects->GetHeadPosition();

	while (pos != NULL)
		((TWorldState*) m_pFactoryObjects->GetNext(pos))->SetFuzzy(val);
}

BOOL TPercept::GetUseFuzzyValues() const
{
	POSITION	pos = m_pFactoryObjects->GetHeadPosition();
	if (pos != NULL)
		return ((TWorldState*) m_pFactoryObjects->GetNext(pos))->GetFuzzy();
	else
		return FALSE;
}

/*-----------------------------------------------------------------------------
Function: void TPercept::SetFuzzyRound(double val)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: value to round fuzzy values at (default 0.5)
Returnval: (void)
Description: sets all worldstates fuzzy round value
-----------------------------------------------------------------------------*/
void TPercept::SetFuzzyRound(double val)
{
	POSITION	pos = m_pFactoryObjects->GetHeadPosition();

	while (pos != NULL)
		((TWorldState*) m_pFactoryObjects->GetNext(pos))->SetFuzzyRound(val);
}

