// GeneticAlgorithm.cpp: implementation of the GeneticAlgorithm class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Basis.h"
#include "GeneticCode.h"
#include "GeneticAlgorithm.h"

#include <time.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

IMPLEMENT_SERIAL( GeneticAlgorithm, CObject, 1 )

GeneticAlgorithm::GeneticAlgorithm()
{
	m_population = new CObList(SIZE_OF_POPULATION);
}

GeneticAlgorithm::~GeneticAlgorithm()
{
	DeletePopulation();
}

void GeneticAlgorithm::DeletePopulation()
{
	if (m_population != NULL)
	{
		while (!m_population->IsEmpty())
			delete m_population->RemoveHead();
		delete m_population;
		m_population = NULL;
	}
}

void GeneticAlgorithm::Serialize( CArchive& ar )
{
	if (ar.IsStoring())
	{
		// save to archive
		ar.WriteObject(m_population);
	}
	else
	{
		// load from archive
		DeletePopulation();
		m_population = (CObList *) ar.ReadObject(NULL);
		GetFirstCodePosition();
	}
}

void GeneticAlgorithm::GenerateNewPopulationFromScratch(GeneticCode* code)
{
	GeneticCode*	newCode;

	DeletePopulation();
	m_population = new CObList(SIZE_OF_POPULATION);

	srand((unsigned) time(NULL));

	// create all new individuals as random distribution from the one passed
	for (int j = 0; j < SIZE_OF_POPULATION; j++)
	{
		newCode = code->CreateRandomCode();
		m_population->AddTail(newCode);
	}
	GetFirstCodePosition();
}

void GeneticAlgorithm::GenerateNewPopulationFromOld()
{
	GeneticCode*	candidates[SIZE_OF_PARENTS];
	double			ratio[SIZE_OF_PARENTS];
	double			sum							= 0.0;
	CObList*		newPopulation				= new CObList(SIZE_OF_POPULATION);
	POSITION		pos;
	int				i;
	double			random;
	int				index1;
	int				index2;

	// sort candidates
	OrderByQuality();

	// copy the candidates
	i = 0;
	pos = m_population->GetHeadPosition();
	while ((pos != NULL) && (i<SIZE_OF_PARENTS))
	{
		candidates[i] = (GeneticCode*) m_population->GetNext(pos);
		sum += candidates[i]->m_quality;
		ratio[i] = sum;
		i++;
	}

	// the best so far is always taken to the next population
	newPopulation->AddTail(new GeneticCode((GeneticCode*)m_population->GetHead()));
	
	// seed random generator
	srand((unsigned) time(NULL));

	for (i = 1; i < SIZE_OF_POPULATION; i++)
	{
		// random number 0 <= random <= 1
		random = sum * (double) rand() / (double) RAND_MAX;

		index1 = 0;
		while ((random > ratio[index1]) && (index1 < SIZE_OF_PARENTS - 1))
			index1++;

		index2 = index1;
		// make sure that no individual is mating with its own :-)
		while (index1 == index2)
		{
			// random number 0 <= random <= 1
			random = sum * (double) rand() / (double) RAND_MAX;

			index2 = 0;
			while ((random > ratio[index2]) && (index2 < SIZE_OF_PARENTS - 1))
				index2++;
		}
		
		// crossover and add to new list
		newPopulation->AddTail(candidates[index1]->CrossOver(candidates[index2]));
	}

	DeletePopulation();
	m_population = newPopulation;
	GetFirstCodePosition();
}

// returns the next code to test, NULL if no more codes
GeneticCode* GeneticAlgorithm::GetNextCode() 
{ 
	if (m_currentCodePosition != NULL)
		return (GeneticCode*) m_population->GetNext(m_currentCodePosition);
	else
		return NULL;
}


// get the code with the best evaluation
GeneticCode* GeneticAlgorithm::OrderAndGetBest()
{
	OrderByQuality();
	return (GeneticCode*)m_population->GetHead();
}

void GeneticAlgorithm::OrderByQuality()
{
	CObList*		sorted = new CObList(SIZE_OF_POPULATION);
	POSITION		pos;
	POSITION		posSorted;
	POSITION		oldPosSorted;
	GeneticCode*	code;
	double			quality;
	BOOL			inserted;

	pos = m_population->GetHeadPosition();
	while (pos != NULL)
	{
		code = (GeneticCode*) m_population->GetNext(pos);

		// insert sorted
		inserted = FALSE;
		posSorted = sorted->GetHeadPosition();
		oldPosSorted = NULL;
		while ((posSorted != NULL) && !inserted)
		{
			quality = ((GeneticCode*)sorted->GetNext(posSorted))->m_quality;
			if (quality < code->m_quality)
			{
				if (oldPosSorted == NULL)
					sorted->AddHead(code);
				else
					sorted->InsertBefore(oldPosSorted, code);
				inserted = TRUE;
			}
			oldPosSorted = posSorted;
		}

		if (!inserted)
			sorted->AddTail(code);
	}

	// delete the old list
	delete m_population;
	m_population = sorted;
}

CString GeneticAlgorithm::ToString()
{
	CString			ret;
	POSITION		pos;
	GeneticCode*	code;

	pos = m_population->GetHeadPosition();
	while (pos != NULL)
	{
		code = (GeneticCode*) m_population->GetNext(pos);
		ret += code->ToString();
	}
	return ret;
}