// TCoachAgent.cpp: implementation of the TCoachAgent class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "stdbnx.h"
#include "Basis.h"
#include "BasisDoc.h"
#include "BasisView.h"
#include "TPerception.h"
#include "TPerceptionParser.h"
#include "GeneticCode.h"
#include "GeneticAlgorithm.h"
#include "TCoachAgent.h"
#include "DGeneticPoolInfo.h"

#include <math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


// settings for debugging coach
// /team living_systems /c /server localhost /port 6001 /rule ruleOfflineCoach.txt
// /team living_systems /c /server 192.168.10.184 /port 6001 /rule ruleOfflineCoach.txt

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

TCoachAgent::TCoachAgent(CBasisDoc* pDoc, CBasisView* pBView, TPerceptionView* pPView
						, TNetworkView* pNView, TRuleView* pRView)
: TDoingAgent(pDoc, pBView, pPView, pNView, pRView)
{
	m_bIsCoach = TRUE;
	m_uiPortAgent = DEFAULT_MASTERCH;
	m_hasEyesOn = FALSE;
	m_hasTypesEvaluated = FALSE;
	m_typeAvailable[0] = 11;
	for (int i = 1; i < NO_OF_PLAYERTYPES; i++)
		m_typeAvailable[i] = 3;

	m_actionStartTime	= 0;
	m_bReadyForAction	= TRUE;
	m_highestQuality	= 0.0;
	m_noOfPlayersChanged = 0;

	// create elements for genetic algorithm
	m_genetics = new GeneticAlgorithm();

	// create a default code
	m_genetics->GenerateNewPopulationFromScratch(m_geneticCode);
	m_currentCode = m_genetics->GetNextCode();
	UseNewCode();

	m_bestCode = m_currentCode;
//	m_bestQualityCurve = new double[100];
	m_populationCount = 0;
	m_bestQuality = 0.0;
	m_trainingResultFile = NULL;
	m_infoDialog = NULL;
}

TCoachAgent::~TCoachAgent()
{
	// delete the goals
	for (int i = 0; i < NO_OF_PLAYERTYPES; i++)
		while (!m_evaluations[i].IsEmpty())
			delete m_evaluations[i].RemoveHead();

	delete m_genetics;
//	delete m_bestQualityCurve;

	if (m_trainingResultFile != NULL)
	{
		m_trainingResultFile->Flush();
		m_trainingResultFile->Close();
		delete m_trainingResultFile;
	}
	if (m_infoDialog != NULL)
		delete m_infoDialog;
}



// called when a message from the socket is received
void TCoachAgent::OnReceivedMessage(LPCTSTR message, BOOL bFromNetwork)
{
	TViewingAgent::OnReceivedMessage(message, bFromNetwork);

	// after see_global, the coach acts
	#ifdef SYNCHRONIZE_ON_SENSE_BODY
		if ((message[1] == 's') && (message[3] == 'e'))
			// it is a see info, sent every 100 ms to a coach
			OnTimer(FALSE);
	#endif

	m_pDoc->SetConnected();
	m_bHadConnectionError = FALSE;
	OnReceivedSee();

	if (!m_hasEyesOn)
	{
		if (SERVER_VERSION < 7)
		{
			SwitchEye(TRUE);
			m_hasEyesOn = TRUE;
			OnAgentReady();
		}

		else if (m_pPerception->HaveAllPlayerTypes())
		{
			int orderOfChange[PLAYERS_PER_TEAM] = {9, 10, 11, 2, 3, 4, 5, 6, 7, 8, 1};

			if (!m_hasTypesEvaluated)
			{
				EvaluatePlayerTypes();
				m_hasTypesEvaluated = TRUE;
			}
			SwitchPlayerType(orderOfChange[m_noOfPlayersChanged]);

			SwitchEye(TRUE);
			if (m_noOfPlayersChanged >= PLAYERS_PER_TEAM - 1)
			{
				m_hasEyesOn = TRUE;
				OnAgentReady();
			}
		}
	}
}

void TCoachAgent::ReceivedTerminate()
{
	OnReceivedLastSenseBody();
}


BOOL TCoachAgent::DoReflexSelection()
{
	return FALSE;
}


void TCoachAgent::EvaluatePlayerTypes()
{
	PlayerTypeEvaluation*	typeEvaluation;

	double*		playerParam = m_pPerception->GetPlayerParams();
	double		evaluation;

	int			parameterIndex[NO_OF_TRADEOFFS]	= {	player_speed_max, 
													player_decay,
													dash_power_rate,
													kickable_margin,
													extra_stamina};

	double*		defaultTypeValues = m_pPerception->GetPlayerType(DEFAULT_PLAYER_TYPE);
	double*		playerType;

	double		defaultValues[NO_OF_TRADEOFFS] = {	defaultTypeValues[player_speed_max],
													defaultTypeValues[player_decay],
													defaultTypeValues[dash_power_rate],
													defaultTypeValues[kickable_margin],
													defaultTypeValues[extra_stamina] };


	double		utilityFactor[PLAYERS_PER_TEAM][NO_OF_TRADEOFFS] = 
				{	
					{0.0, 0.0, 0.0, 0.0, 0.0}, // goalie does not matter
					{0.0, 0.0, 0.0, 0.0, 0.0},
					{0.0, 0.0, 0.0, 0.0, 0.0},
					{0.0, 0.0, 0.0, 0.0, 0.0},
					{0.0, 0.0, 0.0, 0.0, 0.0},
					{0.0, 0.0, 0.0, 0.0, 0.0},
					{0.0, 0.0, 0.0, 0.0, 0.0},
					{0.0, 0.0, 0.0, 0.0, 0.0},
					{1.0, 0.0, 0.0, 0.0, 0.5},
					{1.0, 0.0, 0.0, 0.0, 0.5},
					{1.0, 0.0, 0.0, 0.0, 0.5}
				};

	double		playerCriteriaMin[PLAYERS_PER_TEAM][NO_OF_TRADEOFFS] = 
				{	
					{1.0, 0.4, 0.006, 0.7, 0}, // goalie does not matter
					{1.0, 0.4, 0.006, 0.7, 0},
					{1.0, 0.4, 0.006, 0.7, 0},
					{1.0, 0.4, 0.006, 0.7, 0},
					{1.0, 0.4, 0.006, 0.7, 0},
					{1.0, 0.4, 0.006, 0.7, 0},
					{1.0, 0.4, 0.006, 0.7, 0},
					{1.0, 0.4, 0.006, 0.7, 0},
					{1.0, 0.4, 0.006, 0.7, 0},
					{1.0, 0.4, 0.006, 0.7, 0},
					{1.0, 0.4, 0.006, 0.7, 0}	
				};

	double		playerCriteriaMax[PLAYERS_PER_TEAM][NO_OF_TRADEOFFS] = 
				{	
					{1.2, 0.6, 0.008, 0.9, 100}, // goalie does not matter
					{1.2, 0.6, 0.008, 0.9, 100},
					{1.2, 0.6, 0.008, 0.9, 100},
					{1.2, 0.6, 0.008, 0.9, 100},
					{1.2, 0.6, 0.008, 0.9, 100},
					{1.2, 0.6, 0.008, 0.9, 100},
					{1.2, 0.6, 0.008, 0.9, 100},
					{1.2, 0.6, 0.008, 0.9, 100},
					{1.2, 0.6, 0.008, 0.9, 100},
					{1.2, 0.6, 0.008, 0.9, 100},
					{1.2, 0.6, 0.008, 0.9, 100}	
				};

	BOOL		criteriaAscending[PLAYERS_PER_TEAM][NO_OF_TRADEOFFS] = 
				{	
					{FALSE, FALSE, FALSE, FALSE, FALSE}, // goalie does not matter
					{FALSE, FALSE, FALSE, FALSE, FALSE},
					{TRUE,  FALSE, FALSE, FALSE, FALSE},
					{TRUE,  FALSE, FALSE, FALSE, FALSE},
					{FALSE, FALSE, FALSE, FALSE, FALSE},
					{TRUE,  FALSE, FALSE, FALSE, FALSE},
					{TRUE,  FALSE, FALSE, FALSE, FALSE},
					{TRUE,  FALSE, FALSE, FALSE, FALSE},
					{TRUE,  FALSE, FALSE, FALSE, FALSE},
					{TRUE,  FALSE, FALSE, FALSE, FALSE},
					{TRUE,  FALSE, FALSE, FALSE, FALSE}	
				};


	// loop through all players except goalie
	for (int j = 1; j < PLAYERS_PER_TEAM; j++)
	{
		// look for the best fitting player type for this player
		for (int i = 0; i < NO_OF_PLAYER_TYPES; i++)
		{
			if (i == 0)
				// prefer default type
				evaluation = 0.1;
			else
				evaluation = 0.0;

			playerType = m_pPerception->GetPlayerType(i);

			for (int k = 0; k < NO_OF_TRADEOFFS; k++)
			{
				evaluation += GetLinearFuzzyValue(	playerType[parameterIndex[k]], 
													playerCriteriaMin[j][k], 
													playerCriteriaMax[j][k], 
													criteriaAscending[j][k])
							  * utilityFactor[j][k];
			}

			typeEvaluation = new PlayerTypeEvaluation(i, evaluation);
			InsertSorted(m_evaluations[j], typeEvaluation);
		}
	}
}

void TCoachAgent::SwitchPlayerType(int who)
{
	BOOL		changeDone = FALSE;
	int			optimalType = 0;
	POSITION	pos;

	pos = m_evaluations[who - 1].GetHeadPosition();
	while (!changeDone && (pos != NULL))
	{
		optimalType = ((PlayerTypeEvaluation*) 
					m_evaluations[who - 1].GetNext(pos))->m_type;
		if (m_typeAvailable[optimalType] > 0)
		{
			// Now change the player's type
			CoachChangePlayerType(who, optimalType);
			m_typeAvailable[optimalType]--;
			changeDone = TRUE;
		}
	}

	m_noOfPlayersChanged++;
}

/*-----------------------------------------------------------------------------
Parameter: reference to the list to insert
			  module to insert
Returnval: (void)
Description: inserts the module into the list sorted by the activationexecutability
-----------------------------------------------------------------------------*/
void TCoachAgent::InsertSorted(CObList& list, PlayerTypeEvaluation* module)
{
	POSITION				oldPos;
	POSITION				pos = list.GetHeadPosition();
	PlayerTypeEvaluation*	compare;
	double					value = module->m_value;

	while(pos != NULL)
	{
		oldPos = pos;
		compare = (PlayerTypeEvaluation*) list.GetNext(pos);
		if (value >= compare->m_value)
		{
			list.InsertBefore(oldPos, module);
			return;
		}
	}

	// it is the lowest object so far
	list.AddTail(module);
}

// registers all perceptions of this agent
void TCoachAgent::CreatePerceptions()
{
	m_pBehaviorNetwork->AddPerception("readyForAction", this, (PCallBackPerception) GetReadyForAction);
	m_pBehaviorNetwork->AddPerception("actionFinished", this, (PCallBackPerception) GetActionFinished);
	m_pBehaviorNetwork->AddPerception("maySaySomething", this, (PCallBackPerception) GetMaySaySomething);
	m_pBehaviorNetwork->AddPerception("win", this, (PCallBackPerception) GetWin);
}

// registers all resource perceptions of this agent
void TCoachAgent::CreateResources()
{
	m_pBehaviorNetwork->AddResource("mouth", this, (PCallBackPerception) GetMouthsAvailable, 1);
}
	
// registers all perceptions of this agent
void TCoachAgent::CreateActions()
{
	m_pBehaviorNetwork->AddAction("sayStrategy", this, (PCallBackAction) SayStrategy);
	m_pBehaviorNetwork->AddAction("startAction", this, (PCallBackAction) StartAction);
	m_pBehaviorNetwork->AddAction("evaluateAction", this, (PCallBackAction) EvaluateAction);
	m_pBehaviorNetwork->AddAction("startKick", this, (PCallBackAction) StartKick);
}




// Perceptions
/*-----------------------------------------------------------------------------
true if the coach is on turn to say something
-----------------------------------------------------------------------------*/
double TCoachAgent::GetMaySaySomething()
{
	return 0.0;
}

/*-----------------------------------------------------------------------------
true if the game is won
-----------------------------------------------------------------------------*/
double TCoachAgent::GetWin()
{
	return 0.0;
}


/*-----------------------------------------------------------------------------
offline coach
-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------
true if the training action is over
-----------------------------------------------------------------------------*/
double TCoachAgent::GetActionFinished()
{
	TBall* ball = m_pPerception->GetBall();

	// timeout?
	if ((m_pPerception->GetSeeTime() - m_actionStartTime) > TRAINING_DURATION)
		return 1.0;

	// ball out or in other field?
	if ((ball->GetAbsX() > 1.0) || 
		(ball->GetAbsY() < -PITCH_WIDTH) ||
		(ball->GetAbsY() > PITCH_WIDTH))
		return 1.0;

	return 0.0;
}

/*-----------------------------------------------------------------------------
true if the training action can be started
-----------------------------------------------------------------------------*/
double TCoachAgent::GetReadyForAction()
{
	if (m_bReadyForAction)
		return 1.0;
	else
		return 0.0;
}




// Resources
double TCoachAgent::GetMouthsAvailable()
{
	return 1.0;
}


// Actions

/*-----------------------------------------------------------------------------
turns body to the known ball direction
-----------------------------------------------------------------------------*/
void TCoachAgent::SayStrategy(double activation)
{
}


/*-----------------------------------------------------------------------------
evaluates a training cycle
-----------------------------------------------------------------------------*/
void TCoachAgent::EvaluateAction(double activation)
{

	// calculate speed
	TVector	ballSpeed		= m_pPerception->GetBall()->GetAbsSpeed();
	double	speed			= ballSpeed.GetLength();
	double	qualitySpeed	= GetLinearFuzzyValue(speed, 1.0, BALL_SPEED_MAX);
	double	average;
	int		newActionCount	= m_actionCount + 1;

	// calculate precision
	double	precision = ballSpeed.GetY();
	if (precision < 0.0)
		precision = -precision;
	double	qualityPrecision = GetLinearFuzzyValue(precision, 0.0, 0.2, FALSE);

	// calculate time
	int		time = m_pPerception->GetSeeTime() - m_actionStartTime;
	double	qualityTime = GetLinearFuzzyValue(time, 5, TRAINING_DURATION, FALSE);

	// calculate fitness
	double	quality =	1.0 * qualitySpeed + 
						1.0 * qualityPrecision +
						0.5 * qualityTime;

	// calculate average quality
	average = (m_actionCount * m_averageQuality + quality) / (double) newActionCount;
	
	// standard deviation of quality (incrementally calculated)
	// Sneu = (n+1)(Mneu - M)2 + (n-1)S / n
	if (m_actionCount > 0)
	{
		double deltaMean = average - m_averageQuality;
		m_varianz = newActionCount * (deltaMean * deltaMean) 
										+ ((double)(m_actionCount - 1)
										* m_varianz / (double) m_actionCount);
	}
	m_averageQuality = average;

	if (m_actionCount < TRIES_PER_TYPE)
		// continue with the same settings
		m_actionCount++;
	else
	{
		// enough tested, evaluate the tests and create a new one
		EvaluateCode();
		UseNewCode();
	}

	m_bReadyForAction = TRUE;
	m_bCommandDone = TRUE;
}

/*-----------------------------------------------------------------------------
starts a new training cycle
-----------------------------------------------------------------------------*/
void TCoachAgent::StartAction(double activation)
{
	// move players and ball to their initial position/speed
	CoachMove("(ball)", -5, 0);
	CoachMove("(player living_systems 1)", -13, 0);
	this->CoachChangeMode(PM_PlayOn);
	
	m_actionStartTime = m_pPerception->GetSeeTime();
	m_bReadyForAction = FALSE;
	m_bCommandDone = TRUE;
}

/*-----------------------------------------------------------------------------
starts a new training cycle for kicking fast ball
-----------------------------------------------------------------------------*/
void TCoachAgent::StartKick(double activation)
{
	int		angle = (m_actionCount * 10) % 360;
	// only kick from the front
	if ((angle > 90) && (angle < 270))
		angle = 180 - angle;

	int		px = -15;
	int		py = 0;
	double	distance = 15.0;
	int		x = px + (int) (cos(angle * DEG2RAD) * distance);
	int		y = py + (int) (sin(angle * DEG2RAD) * distance);
	TVector	speed = TVector(px, py) - TVector(x, y);
	speed.Normalize();
	speed *= BALL_SPEED_MAX;

	// move players and ball to their initial position/speed
	CoachMove("(ball)", x, y, 0, 0, 0);
	CoachMove("(player living_systems 1)", px, py, angle, 0, 0);
	for (int i = 0; i < 4; i++)
	{
		CoachMove("(ball)", x, y, 0, 0, 0, TRUE);
		CoachMove("(player living_systems 1)", px, py, angle, 0, 0, TRUE);
	}
	CoachMove("(ball)", x, y, 0, speed.GetX(), speed.GetY(), TRUE);
	CoachMove("(player living_systems 1)", px, py, angle, 0, 0, TRUE);
	
	this->CoachChangeMode(PM_PlayOn);
	
	m_actionStartTime = m_pPerception->GetSeeTime();
	m_bReadyForAction = FALSE;
	m_bCommandDone = TRUE;
}


/*-----------------------------------------------------------------------------
called once a complete training cycle with one agent is finished
-----------------------------------------------------------------------------*/
void TCoachAgent::EvaluateCode()
{
	if (m_averageQuality > m_highestQuality)
		m_highestQuality = m_averageQuality;
	
	// a test is over so now tell the genetic code how good it was
	m_currentCode->m_quality = m_averageQuality;
	m_currentCode->m_varianz = m_varianz;

	// and go to the next code
	m_currentCode = m_genetics->GetNextCode();

	// update the coach info dialogs data
	InfoDialogUpdate();

	if (m_currentCode == NULL)
	{
		// all individuals had their time
		CString text;
		GeneticCode* best = m_genetics->OrderAndGetBest();
//		m_bestQualityCurve[m_populationCount] = best->m_quality;
		m_populationCount++;
		if (m_bestQuality < best->m_quality)
		{
			m_bestQuality = best->m_quality;
			m_bestCode = best;
		}

		// write the best code to the training log
		text.Format("%d %s", m_populationCount, best->ToString());
		WriteToTrainingLog(text);

		// all Codes have been tested, so setup a new population
		m_genetics->GenerateNewPopulationFromOld();
		m_currentCode = m_genetics->GetNextCode();
	}
}

/*-----------------------------------------------------------------------------
called to use a new genetic code
-----------------------------------------------------------------------------*/
void TCoachAgent::UseNewCode()
{
	// tell the player its new code
	TellCode();

	m_averageQuality = 0.0;
	m_varianz = 0.0;
	m_actionCount = 0;
}

/*-----------------------------------------------------------------------------
creates and sends a coach say message to inform the agent about the parameters
-----------------------------------------------------------------------------*/
void TCoachAgent::TellCode()
{
	CString msg;

	// pass the side identifier, receiver, sender, stamina and Genetic idenitfier G
	msg.Format("%s " , GetSideIdentifier());
	msg += m_currentCode->ToString();
	
	Say(msg);
}

/*-----------------------------------------------------------------------------
writes the passed string into the result logfile of the training
-----------------------------------------------------------------------------*/
void TCoachAgent::WriteToTrainingLog(CString buffer)
{
	UINT	mode;
	CString filename;

	// if the file is not already open, open it
	if (m_trainingResultFile == NULL)
	{
		m_trainingResultFile = new CFile();
		mode = CFile::modeCreate | CFile::modeWrite;
	}
	else
		mode = CFile::modeWrite;

	
	filename.Format("TrainingResults%d.txt", this->GetPort());
	if(!m_trainingResultFile->Open(filename, mode))
	{
		AfxMessageBox("Unable to open training result file for writing");
		delete m_trainingResultFile;
		m_trainingResultFile = NULL;
		return;
	}

	// write the last result
	m_trainingResultFile->SeekToEnd();
	m_trainingResultFile->Write(buffer, buffer.GetLength());
	m_trainingResultFile->Flush();
	m_trainingResultFile->Close();
}

void TCoachAgent::ShowGeneticPoolInfo()
{
	if (m_infoDialog == NULL)
	{
		m_infoDialog = new DGeneticPoolInfo();
	}

	m_infoDialog->ShowWindow(TRUE);
	InfoDialogUpdate();
}

void TCoachAgent::InfoDialogUpdate()
{
	if ((m_infoDialog != NULL) && (m_infoDialog->IsWindowVisible()))
	{
		m_infoDialog->m_content = GetGeneticPoolInfo();
		m_infoDialog->m_noOfGenerations = m_populationCount;

		m_infoDialog->UpdateData(FALSE);
	}
}

CString	TCoachAgent::GetGeneticPoolInfo()
{
	return m_genetics->ToString();
}

