// TDoingAgent.cpp: implementation of the TDoingAgent class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "stdbnx.h"
#include "Basis.h"
#include "TVector.h"
#include "BasisDoc.h"
#include "BasisView.h"
#include "TPerception.h"
#include "TThisPlayer.h"
#include "GeneticCode.h"
#include "TDoingAgent.h"
#include <math.h>

// minimal dash power (for action parametrization) min 0 max 100
#define MIN_DASH					60
#define MAX_SELECTION_TRIES			5
#define DIFF_ANGLE					110

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*-----------------------------------------------------------------------------
Function: TDoingAgent::TDoingAgent(CBasisDoc* pDoc, CBasisView* pBView, TPerceptionView* pPView
								 , TNetworkView* pNView, TRuleView* pRView) 
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: pointer to the document
			  pointer to the view containing network traffic
			  pointer to the view for perception visualization
			  pointer to the view for network visualization
			  pointer to the view for behavior rules
Returnval: (void)
Description: constructor
-----------------------------------------------------------------------------*/
TDoingAgent::TDoingAgent(CBasisDoc* pDoc, CBasisView* pBView, TPerceptionView* pPView
								 , TNetworkView* pNView, TRuleView* pRView) 
		: TViewingAgent(pBView, pPView, pNView, pRView)
{
	m_pDoc = pDoc;
	m_bAgentIsReady = FALSE;
	m_bReceivedSee = FALSE;
	m_bSelectedReflex = FALSE;
	m_iSenseBodyCount = 0;
	
	// for statistics
	m_dAvgDistance = 0.0;
	m_lNoOfRelaxes = 0L;
	m_iNoOfCycles = 0;
	m_iNoOfConcurrentCycles = 0;
	m_bWantNarrowView = FALSE;
	m_bWantWideView = FALSE;
	m_iNoOfLazyCycles = 0;
	m_expectedSee = 0;
/*
	// generate the parameters that control this agent
	double parameters[SIZE_OF_CODE] =
	{
		FIRST_KICK_POWER,
		SECOND_KICK_POWER,
		MIN_KICK_ANGLE,
		SECURE_DRIBBLE,
		DRIBBLE_ANGLE,
		CLOSE_DRIBBLE_POWER
	};

	double range[SIZE_OF_CODE] =
	{
		FIRST_KICK_POWER_RANGE,
		SECOND_KICK_POWER_RANGE,
		MIN_KICK_ANGLE_RANGE,
		SECURE_DRIBBLE_RANGE,
		DRIBBLE_ANGLE_RANGE,
		CLOSE_DRIBBLE_POWER_RANGE
	};
*/

	double parameters[SIZE_OF_CODE] =
	{
		MIN_BALL_SPEED,
		STOP_KICK_POWER,
		MIN_TURN_DIR,
	};

	double range[SIZE_OF_CODE] =
	{
		MIN_BALL_SPEED_RANGE,
		STOP_KICK_POWER_RANGE,
		MIN_TURN_DIR_RANGE,
	};

	m_geneticCode = new GeneticCode(0.0, parameters, range);
}

TDoingAgent::~TDoingAgent()
{
	delete m_geneticCode;
}

void TDoingAgent::SetGeneticCode(TParseString& codeString)
{
	m_geneticCode->FromString(codeString);

	CString msg;
	msg.Format("Received Parameters: %s ", m_geneticCode->ToString());
	m_pView->DisplayDebug(msg, KATEGORY_PERCEPTION, DEBUG_MIDPERCEPTION);
}


/*-----------------------------------------------------------------------------
Function: void TDoingAgent::SetKnowledgeType(int type)
Date    : 19.11.98
-------------------------------------------------------------------------------
Parameter: type of knowledge (MAES or DORER or DORER2)
Returnval: (void)
Description: calls base class function and adds action functions
-----------------------------------------------------------------------------*/
void TDoingAgent::SetKnowledgeType(int type)
{
	// create the network
	TViewingAgent::SetKnowledgeType(type);

	// add the perception functions
	CreateActions();
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::OnReceivedMessage(LPCTSTR message)
Date    : 29.01.98
-------------------------------------------------------------------------------
Parameter: pointer to the message string received
Returnval: (void)
Description: calls the base class function to parse the message and then calls
				 to update the percept class in knowledge
-----------------------------------------------------------------------------*/
void TDoingAgent::OnReceivedMessage(LPCTSTR message, BOOL bFromNetwork)
{
	if ((message[1] == 'e') && (message[2] == 'r'))
	{
		// an error message is here
//		m_pDoc->OnConnectionError();
		m_bHadConnectionError = TRUE;
	}
	else
	{
		// tell the document, that we have connection
		m_pDoc->SetConnected();
		m_bHadConnectionError = FALSE;
	}

	TViewingAgent::OnReceivedMessage(message, bFromNetwork);

	// update the propositions in knowledge
//	m_pBehaviorNetwork->UpdatePerception(bFromNetwork);

	// after sense body, the agent acts
	#ifdef SYNCHRONIZE_ON_SENSE_BODY
		if ((message[1] == 's') && (message[3] == 'n'))
		{
			// it is a sensebody info, sent every 100 ms (new cycle)
			m_bHaveActedThisCycle = FALSE;
			OnTimer(FALSE);
		}
		else if ((message[1] == 's') && (message[3] == 'e') && (m_expectedSee <= 0))
		{
			// it is a sensebody info, sent every 100 ms
			if (!m_bHaveActedThisCycle)
				OnTimer(TRUE);
		}
	#endif
}

// for autorunning games

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::OnReceivedSee()
Date    : 23.06.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: called, when the first see command is perceived. calls the documents
				 AutorunReceivedSee, to tell the soccermonitor, the agent is alive
-----------------------------------------------------------------------------*/
void TDoingAgent::OnReceivedSee()
{
	if (!m_bReceivedSee)
	{
		m_bReceivedSee = TRUE;
		if (m_bAgentIsReady)
			m_pDoc->AutorunAgentReady();
	}
}

void TDoingAgent::OnAgentReady()
{
	if (!m_bAgentIsReady)
	{
		m_bAgentIsReady = TRUE;
		if (m_bReceivedSee)
			m_pDoc->AutorunAgentReady();
	}
}

void TDoingAgent::OnReceivedLastSenseBody()
{
	m_pDoc->AutorunSendResults();
}


void TDoingAgent::ReceivedTerminate()
{
	//SenseBody();
	m_pPerception->SetWaitForLastSenseBody(TRUE);
}

BOOL TDoingAgent::IsAutorun() const
{
	return m_pDoc->IsAutorun();
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::OnTimer(BOOL bActOnSee)
Date    : 23.03.98
-------------------------------------------------------------------------------
Parameter: true if called by a see message, false if called by a sense body message
Returnval: (void)
Description: called every 100 ms to calculate the best suitable action
-----------------------------------------------------------------------------*/
void TDoingAgent::OnTimer(BOOL bActOnSee)
{
	int	iNoOfTrys;
	BOOL	bReady;

#ifdef SHOW_TIME
	
	LARGE_INTEGER	frequency;
	LARGE_INTEGER	start;
	LARGE_INTEGER	stop;
	double			time;
		
	if (QueryPerformanceFrequency(&frequency))
		QueryPerformanceCounter(&start);
#endif

	// time has proceeded
	m_bSelectedReflex = FALSE;
	
	// perception is updated in the case of no new real perception
	if (!bActOnSee)
		m_pPerception->UpdatePerception();
	
	// future ball positions have to be calculated each timeslice
	m_pPerception->CalculateFutureBallPositions();

	// only calculate something, if the agent is connected to the server
	if (!IsConnected())
	{
		// update the perception propositions of the behavior network
		// if the agent is not connected, the network is probably a test network,
		// so e.g. true and false are calculated (without perception)
		m_pBehaviorNetwork->UpdatePerception(FALSE);

		// activation is spread at least once
		m_pBehaviorNetwork->DoActivationSpreading();

		m_pNetworkView->RedrawView();
		return;
	}

	// only do action if the agent is running
	if (m_pView->GetRunning())
	{
		// selection of reflexes
		if (DoReflexSelection())
		{
			m_pView->DisplayDebug("Did reflex selection", KATEGORY_ACTION, DEBUG_HIGHACTION);
			// a reflex has been executed, so no further action to select
			return;
		}

		// check if we get see info every 75ms 
		// if this is the case, we act on see not on sense body
		if (m_expectedSee > 0)
		{
			m_pView->DisplayDebug("Will act on see", KATEGORY_ACTION, DEBUG_LOWACTION);
			return;
		}

		// for debugging
		DisplayBallPosition(m_pPerception->GetBall()->GetPosition(), m_pPerception->GetBall()->GetAbsSpeed(), FALSE);
		TThisPlayer* agent = m_pPerception->GetThisPlayer();
		DisplayPosition(agent->GetPosition(), agent->GetAbsBodyDir()
						, agent->GetAbsDir(), FALSE);

		// execution of queued actions
		if (ExecuteQueuedAction())
			// a queued action has been executed, so no further actions
			return;

		// the agent has to wait for the next see perception
		if (GetWaitForNextSee())
			return;

		// update the perception propositions of the behavior network
		m_pBehaviorNetwork->UpdatePerception(TRUE);

		// activation is spread at least once
		m_pBehaviorNetwork->DoActivationSpreading();

		// repeat until an action has been selected
		iNoOfTrys = 0;
		bReady = FALSE;
		m_bCommandDone = FALSE;
		m_bSayDone = FALSE;
		m_bWantNarrowView = FALSE;
		m_bWantWideView = FALSE;
		m_bHaveTurned = FALSE;
		while (!bReady && (iNoOfTrys < MAX_SELECTION_TRIES))
		{
			bReady = m_pBehaviorNetwork->DoActionSelection();
			if (!bReady)
			{
				// reduce the threshold
				m_pBehaviorNetwork->ReduceThreshold();
				// and do a further activation spreading
				m_pBehaviorNetwork->DoActivationSpreading();
			}
			else if (!m_bCommandDone)
			{
				bReady = FALSE;
				// reduce the threshold
				m_pBehaviorNetwork->ReduceThreshold();
				// and do a further activation spreading
				m_pBehaviorNetwork->DoActivationSpreading();
			}
			iNoOfTrys++;
		}
	}
	m_bHaveActedThisCycle = TRUE;

	// redraw network, if it is drawn only after a complete spreading
	if (m_pNetworkView->GetDrawAtEndOfActionSelection())
		m_pNetworkView->RedrawView();

#ifdef SHOW_TIME

	if (QueryPerformanceCounter(&stop))
	{
		time = (double)(stop.LowPart - start.LowPart) / (double) frequency.LowPart;
		TRACE1("acttime: %4.3f\n", time);
	}
#endif
}

/*-----------------------------------------------------------------------------
Function: BOOL TBehaviorNetwork::DoReflexSelection()
Date    : 16.03.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: TRUE, if an action has been performed
Description: checks the situation for reflexes to be done, such as moving to
				 own half after a goal
-----------------------------------------------------------------------------*/
BOOL TDoingAgent::DoReflexSelection()
{
	int	iPlayMode = m_pPerception->GetPlayMode();

	switch (iPlayMode)
	{
		case BEFORE_KICK_OFF:
			// before kickoff, the player moves to his position
			{
				int	x = GetPosKickOffLeft().x;
				int	y = GetPosKickOffLeft().y;
				int	x2 = x;
				int	y2 = y;
				
				// this test is needed because of a strange behavior of the server's move
				// command: players have to be moved with negative x coordinate, no matter
				// on which side they are playing. The GetIsAt function on the other side uses 
				// the 'real' coordinates of a player and therefor players playing from left to 
				// right have to negate the x coordinate for testing (but not for moving)
				if (GetSide() == RIGHT_TO_LEFT)
				{
					x2 = -x2;
					y2 = -y2;
				}

				if (!m_pPerception->GetIsAt(x2, y2))
				{
					Move(x, y);
					
					// perception gets initialized, because object persistence is no longer
					// valid. Ball and players move far distances in no time
					m_pPerception->Initialize();
				}
				else
				{
					// the agent is ready
					OnAgentReady();
					LookToBall();
				}
			}
			return TRUE;	// no more actions have to be taken
/*		
		case KICK_OFF_L:
		case KICK_IN_L:
		case GOAL_KICK_L:
		case FREE_KICK_L:
		case CORNER_KICK_L:
			// if this is a right team player, and is on position, it can wait
			// until game is play on
			// may be should be not dealt with as reflex
			if (GetSide() != LEFT_TO_RIGHT)
			{
				if (m_pPerception->GetObjectDistance(m_pPerception->GetBall()) < 11.0)
				{
					// the player has to stay 9 m away of ball
					LookToBall();
					return TRUE;
				}
			}
			break;

		case KICK_OFF_R:
		case KICK_IN_R:
		case GOAL_KICK_R:
		case FREE_KICK_R:
		case CORNER_KICK_R:
			// if this is a right team player, and is on position, it can wait
			// until game is play on
			// may be should be not dealt with as reflex
			if (GetSide() == LEFT_TO_RIGHT)
			{
				if (m_pPerception->GetObjectDistance(m_pPerception->GetBall()) < 11.0)
				{
					// the player has to stay 9 m away of ball
					DoSearchBall();
					return TRUE;
				}
			}
			break;
*/
		case GOAL_R:
			// after a goal, the player moves to his position
			{
				int	x = GetPosKickOffLeft().x;
				int	y = GetPosKickOffLeft().y;
				
				if (!m_pPerception->GetIsAt(x, y))
					Move(x, y);
				else
					DoSearchBall();
			}
			return TRUE;	// no more actions have to be taken
		
		case GOAL_L:
			// after a goal, the player moves to his position
			{
				int	x = GetPosKickOffRight().x;
				int	y = GetPosKickOffRight().y;
				
				if (!m_pPerception->GetIsAt(x, y))
					Move(x, y);
				else
					LookToBall();
			}
			return TRUE;	// no more actions have to be taken

		case OFFSIDE_L:
		case OFFSIDE_R:
			break;
	}

	
	// if the ball is close, we have narrow view mode
	double distance = VIEW_NEIGHBORHOOD;

	if (m_pPerception->GetObjectDistance(m_pPerception->GetBall()) < distance)
	{
		if (m_pPerception->GetViewAngle() == NORMAL_VIEW)
			ChangeView(NARROW_VIEW, 1);
	}
	else
	{
		if (m_pPerception->GetThisPlayer()->GetNum() == GOALIE_ID)
		{
			// the goalie has always narrow view if ball is near goal
			double dist = m_pPerception->GetObjectDistance(m_pPerception->GetBall());
			if ((dist < 25.0) && (m_pPerception->GetViewAngle() != NARROW_VIEW))
				ChangeView(NARROW_VIEW, 1);
			else if ((dist >= 25.0) && (m_pPerception->GetViewAngle() != NORMAL_VIEW))
				ChangeView(NORMAL_VIEW, 1);
		}
		else if ((m_pPerception->GetViewAngle() == NARROW_VIEW) && !m_bWantNarrowView)
			ChangeView(NORMAL_VIEW, 1);
		else if ((m_pPerception->GetViewAngle() == WIDE_VIEW) && !m_bWantWideView)
			ChangeView(NORMAL_VIEW, 1);
	}
	return FALSE;
}

/*-----------------------------------------------------------------------------
Function: Functions passed to the Behavior network
Date    : 08.03.99
-----------------------------------------------------------------------------*/
void TDoingAgent::SetGoalTrackingOn(BOOL bVal)
{
	if (m_iKnowledgeType == DORER_KNOWLEDGE)
		((TDorerNetwork*) m_pBehaviorNetwork)->SetGoalTrackingOn(bVal);
}

int TDoingAgent::GetGoalTrackingOn() const
{
	if (m_iKnowledgeType == DORER_KNOWLEDGE)
		return ((TDorerNetwork*) m_pBehaviorNetwork)->GetGoalTrackingOn();
	else
		return FALSE;
}

void TDoingAgent::SetUseFuzzyValues(BOOL bVal)
{
	m_pBehaviorNetwork->SetUseFuzzyValues(bVal);
}

int TDoingAgent::GetUseFuzzyValues() const
{
	return m_pBehaviorNetwork->GetUseFuzzyValues();
}

void TDoingAgent::SetFuzzyRound(double dVal)
{
	m_pBehaviorNetwork->SetFuzzyRound(dVal);
}

void TDoingAgent::SetAdaptActivationAreaOn(BOOL bVal)
{
	if (m_iKnowledgeType == DORER_KNOWLEDGE)
		((TDorerNetwork*) m_pBehaviorNetwork)->SetAdaptActivationAreaOn(bVal);
}

int TDoingAgent::GetAdaptActivationAreaOn() const
{
	if (m_iKnowledgeType == DORER_KNOWLEDGE)
		return ((TDorerNetwork*) m_pBehaviorNetwork)->GetAdaptActivationAreaOn();
	else
		return FALSE;
}

void TDoingAgent::SetCombo(CComboBox* pCombo, BOOL bWhich)
{
	m_pBehaviorNetwork->SetCombo(pCombo, bWhich);
}

CString TDoingAgent::GetActionPerformedToString()
{
	// returns the last performed action, once the agent is ready
	if (!m_bAgentIsReady)
		return CString("Not Ready");
	else
		return m_pBehaviorNetwork->GetActionPerformedToString();
}

CString TDoingAgent::GetResultsInitial() const
{
	CString	content;
	content = " NotExecuted vMax Decay DPR Margin ExtraSta";
/*
	content += m_pBehaviorNetwork->GetResultsInitial();
*/
	return content;
}

CString TDoingAgent::GetResults()
{
	CString	result;
	result.Format(" %5d %3.2f %3.2f %4.3f %3.2f %3d n", 
		m_pPerception->GetNoOfCommandsNotExecuted(),
		m_pPerception->GetPlayerSpeedMax(),
		m_pPerception->GetPlayerDecay(),
		m_pPerception->GetDashPowerRate(),
		m_pPerception->GetKickableMargin(),
		m_pPerception->GetExtraStamina());
	
/*
	// get the results from knowledge
	result += m_pBehaviorNetwork->GetResults();
*/
	return result;
}

// for blocksworld
BOOL TDoingAgent::IsBlocksworld()
{
	return m_pBehaviorNetwork->IsBlocksworld();
}

BOOL TDoingAgent::PerformCycle(TNetworkParams* params, int iNoOfCycles)
{
	return m_pBehaviorNetwork->PerformCycle(params, iNoOfCycles);
}

void TDoingAgent::StopAgent()
{
	m_pDoc->StopAgent();
}

// higher level actions

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::TurnNeck(int iDir, BOOL bQueue)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: angle to turn neck
			  flag indicating whether the action should be queued or immediately executed (default)
Returnval: (void)
Description: turns the neck of the agent, but only if turn angle is more than 
				 a threshold
-----------------------------------------------------------------------------*/
void TDoingAgent::TurnNeck(int iDir, BOOL bQueue)
{
	// check if head is at maxturn
//	TDirection headAngle = m_pPerception->GetHeadAngle();
//	TDirection turnDir = iDir + headAngle;
	int headAngle = m_pPerception->GetHeadAngle();
	int turnDir = iDir + headAngle;
	
	if (turnDir < -90)
		turnDir = -90 - headAngle;
	else if (turnDir > 90)
		turnDir = 90 - headAngle;
	else
		turnDir = iDir;

	// do not turn neck if angle is too small
	if ((turnDir < -4) || (turnDir > 4))
		CBaseAgent::TurnNeck(turnDir, bQueue);
}




/*-----------------------------------------------------------------------------
Parameter: none
Returnval: new position of defender or midfielder
Description: 
-----------------------------------------------------------------------------*/
CPoint TDoingAgent::CalculateDefendingPosition()
{
	// going back by defenders
	int			BACKMOST_DEFENDER = -25;
	int			FRONTMOST_MIDFIELDER = 30;
	int			x			= GetPosHome().x;
	int			y			= GetPosHome().y;
	TPlayer*	pOpp		= m_pPerception->GetOpponentToMark();
	TPlayer*	teammate	= m_pPerception->Get2BackmostTeammate();
	TPlayer*	defender	= m_pPerception->Get2BackmostOpponent();
	int			oppX;
	int			oppY;
	int			xOffside;
	int			xOffsideOther;
	int			xBall = (int) m_pPerception->GetFutureBallPosition(15)->GetX();
	int			xBallNow = (int) m_pPerception->GetFutureBallPosition(1)->GetX();
	int			yBallNow = (int) m_pPerception->GetFutureBallPosition(1)->GetY();
	int			x2 = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int			number = m_pPerception->GetThisPlayer()->GetNum();
	CPoint		newPos;
	CString		msg;
	int			newX;
	int			newY;

	// get the own offside line
	if (teammate != NULL)
	{
		xOffside = (int) teammate->GetAbsX();
		msg.Format("2backmostTeam: x:%d", xOffside);
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
	}
	else 
		xOffside = xBall;

	// get the other offside line
	if (defender != NULL)
		xOffsideOther = (int) defender->GetAbsX();
	else 
		xOffsideOther = xBallNow;

	// make coordinates side safe
	if (GetSide() == RIGHT_TO_LEFT)
	{
		xOffside = -xOffside;
		xOffsideOther = -xOffsideOther;
		xBall = -xBall;
		xBallNow = -xBallNow;
		yBallNow = -yBallNow;
		x2 = -x2;
	}

	if (xBall < xOffside + 10)
		// if the ball is closer to the own goal than our defenders, push backwards
		xOffside = xBall - 10;
	else if (teammate != NULL)
	{
		// offside line is 4 m in front of backmost teammate
		int deltaTime = m_pPerception->GetSeeTime() - teammate->GetTime();
		if (deltaTime > 8)
			deltaTime = 8;
		xOffside += 3 + deltaTime;
	}

	// but do not move offside line too far back
	if (xOffside < -47)
		xOffside = -47;

	if (xOffside > 0)
		xOffside = 0;

	if (xOffsideOther < 0)
		xOffsideOther = 0;

	if (xBallNow > xOffsideOther)
		xOffsideOther = xBallNow;

	if ((pOpp != NULL) && (number <= MAX_MIDFIELDER))
	{
		// we have to mark an opponent
		oppX = (int) pOpp->GetAbsX();
		oppY = (int) pOpp->GetAbsY();
		if (GetSide() == RIGHT_TO_LEFT)
		{
			oppX = -oppX;
			oppY = -oppY;
		}

		// if the ball is very near own goal and playmode not playon we move behind opp
		newX = xOffside;
		
		if ((xBall < -35) && (m_pPerception->GetPlayMode() != PLAY_ON))
			newX = oppX - 4;

		// only mark if it is not in offside
		else if (oppX > xOffside)
		{
			if (oppX < x2)
				// opponent is closer to own goal, so try to run behind
				newX = oppX - 8;
			else if ((oppX > -25) && (xBall - oppX < 10))
				// in midfield and if ball is close stand far behind
				newX = oppX - 5;
			else
				// near goal stay closer to opponent
				newX = oppX - 3;

			// at the midline midfielders do not mark if ball is in front
			if ((number > MAX_DEFENDER) && (newX > -10) && (xBall > 20))
				newX = xBall - 20;
		}

		if (((m_pPerception->GetPlaymodeOffsideOther()) ||
			(m_pPerception->GetPlaymodeFreekickOwn())))
		{
			// if it is a goalie kick, we move forward
			if (m_pPerception->GetOwnPlayer(0)->GetDistanceTo(m_pPerception->GetBall()) < 1.0)
			{
				if (number <= MAX_DEFENDER)
				{
					if (newX < BACKMOST_DEFENDER)
						newX = BACKMOST_DEFENDER;
				}
				else
				{
					if (newX < -25)
						newX = -25;
				}
			}
			else if (newX < xBall)
			{
				// in case we have a freekick
				if (number <= MAX_DEFENDER)
				{
					newX = xBall;
					if (newX < BACKMOST_DEFENDER)
						newX = BACKMOST_DEFENDER;
					if (newX > 0)
						newX = 0;
				}
				else
				{
					newX = xBall + 20;
					if (newX < -25)
						newX = -25;
					if (newX > xOffsideOther - 10)
						newX = xOffsideOther - 10;
				}
			}
			msg.Format("freekick: move to x:%d", newX);
			m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		}

		int	iPlayMode = m_pPerception->GetPlayMode();
		if ((newX > -1) && ((iPlayMode == KICK_OFF_L) || (iPlayMode == KICK_OFF_R)))
			newX = -1;

		//  calculate y coordinate
		if (yBallNow < -10)
			if (oppY > 5)
				newY = oppY - 2;
			else
				newY = oppY - 1;
		else if (yBallNow < 10)
			newY = oppY;
		else
			if (oppY < -5)
				newY = oppY + 2;
			else
				newY = oppY + 1;

		if (GetSide() == RIGHT_TO_LEFT)
			newPos = CPoint(-newX , -newY);
		else
			newPos = CPoint(newX , newY);

		msg.Format("Mark Opponent %d at (%d, %d)", pOpp->GetNum(), newX, newY);
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
	}
	else
	{
		// there is no opponent to hide so move to height of ball
		newY = GetPosY();
		if (number <= MAX_DEFENDER)
		{
			// defenders move to height of ball
			if (xOffside >= -1)
				newX = 0;
			else
				newX = xOffside - 1;

			if (newX > 0)
				// defenders move just to midline
				newX = 0;

			if (((m_pPerception->GetPlaymodeOffsideOther()) ||
				(m_pPerception->GetPlaymodeFreekickOwn())))
			{
				if ((m_pPerception->GetOwnPlayer(0)->GetDistanceTo(m_pPerception->GetBall()) < 1.0)
					|| (newX < xBall))
				{
					newX = xBall;
					if (newX < BACKMOST_DEFENDER)
						newX = BACKMOST_DEFENDER;
					if (newX > 0)
						newX = 0;
				}
				msg.Format("Offside of other team: move back to %d", newX);
				m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
			}
		}
		else if (number <= MAX_MIDFIELDER)
		{
			// midfielder keeps 10m before defender or ball
			if (xOffside < -10)
				newX = xOffside + 15;
			else
				newX = xOffside + 25;

			if (newX < -25)
				// midfielder maximally push back to x = -20
				newX = -25;
			
			// stay a little more back if others have a freekick
			if (m_pPerception->GetPlaymodeFreekickOther() && (newX > FRONTMOST_MIDFIELDER - 10))
				newX = FRONTMOST_MIDFIELDER - 10;

			if (newX > FRONTMOST_MIDFIELDER)
				// midfielder maximally push back to x = -20
				newX = FRONTMOST_MIDFIELDER;

			if (newX > xOffsideOther - 10)
				newX = xOffsideOther - 10;
		}
		else
		{
			// offender
			if (xBallNow >= xOffsideOther)
				newX = xOffsideOther - 2;
			else
				newX = xOffsideOther - 4;
			if ((newX > 30) && m_pPerception->GetPlaymodeFreekickOther())
				// if the others have a freekick we do not move forward
				newX = 30;

			// do not run too deep into own half
			if (newX < -3)
				newX = -3;

			if (newX > 41)
				newX = 41;

			// do not stand in other half if ball is near own goal
			if (newX - xBall > 50)
				newX = xBall + 50;

			// move slightly to the side of the ball if it is on the outer side
			if (xBallNow > 30)
			{
				if (yBallNow < -15)
					newY -= 5;
				else if (yBallNow > 15)
					newY += 5;
			}
		}
		
		int	iPlayMode = m_pPerception->GetPlayMode();
		if ((newX > -1) && ((iPlayMode == KICK_OFF_L) || (iPlayMode == KICK_OFF_R)))
			newX = -1;

		if (GetSide() == RIGHT_TO_LEFT)
			// no -newY since GetPosY() already does this
			newPos = CPoint(-newX , newY);
		else
			newPos = CPoint(newX , newY);

		msg.Format("My Coordinates (%d, %d) offside:%d offsideOther:%d", newX, newY, xOffside, xOffsideOther);
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
	}	

	return newPos;
}


/*-----------------------------------------------------------------------------
Function: void TDoingAgent::ComplexDash(double dist, int iDir)
Date    : 24.02.98
-------------------------------------------------------------------------------
Parameter: distance of the goal
			  relative direction of the goal, where to dash
			  power with which to dash (default is 0, which means, that the power is
			  calculated from the distance
			  flag indicating, whether the player should look at the ball while running
Returnval: (void)
Description: avoids to dash into another agent
				 dashes backward, if it is faster (for short traces)
-----------------------------------------------------------------------------*/
void TDoingAgent::ComplexDash(double dist, int iDir, int iPower, BOOL bMindBall, BOOL checkSourrounding)
{
	TDirection	dir = iDir;
	TDirection	goalDir;
	TDirection	deltaDir;
	TDirection	deltaGoalDir;
	int			minDir;
	int			opponentDir;
	double		minDist = 2.0 * m_pPerception->GetPlayerSize();
	TDirection	ballDir;
	TDirection	deltaBallDir;
	TDirection	headAngle = m_pPerception->GetHeadAngle();
	TDirection	actualTurnDir;
	TPlayer*	pPlayer = m_pPerception->GetPlayerInWay(iDir, 5.0);
	CString		msg;
	TThisPlayer* pAgent = m_pPerception->GetThisPlayer();
	BOOL		willTurn;
	int			viewAngle = m_pPerception->GetViewAngle();
	BOOL		runLeft;

	msg.Format("ComplexDash power: %d dir: %d dist:%4.2f", iPower, iDir, dist);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_MIDACTION);

	// if power is not set explicitly, the distance defines the dash power
	if (iPower == 0)
	{
		if (GetOrdinaryGame())
			iPower = 100;
		else 
			iPower = (int) (100.0 * GetLinearFuzzyValue(dist, 0.0, 5.0));
	}

	// avoid to run into other players
	if (pPlayer != NULL)
	{
		double distPlayer = m_pPerception->GetObjectDistance(pPlayer);
		if ((distPlayer < dist) && (pPlayer->IsValid() || (distPlayer > VIEW_NEIGHBORHOOD)))
		{
			// there is a player in way, so try to face other direction
			opponentDir		= pAgent->GetAngleToBody(pPlayer->GetPosition());
			deltaDir		= iDir - opponentDir;
			goalDir			= pAgent->GetAngleToBody(m_pPerception->GetOwnGoal()->GetPosition());
			deltaGoalDir	= iDir - goalDir;

			// check if we are in own half and running backwards.
			if (pAgent->IsInOwnHalfOfField() && (deltaGoalDir > -90) && (deltaGoalDir < 90))
			{
				// in that case we always avoid an object by running closer to the goal
				if (deltaGoalDir > 0)
					runLeft = TRUE;
				else
					runLeft = FALSE;
			}
			else if (deltaDir < 0)
				runLeft = TRUE;
			else
				runLeft = FALSE;


			msg.Format("ComplexDash avoid object: oppenentDir %d distPlayer: %4.2f deltaDir:%d"
						, opponentDir, distPlayer, deltaDir);
			m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
			
			if (distPlayer < 0.7)
				if (runLeft)
					dir = iDir - 90;
				else
					dir = iDir + 90;
			else if (distPlayer < 1.2)
				if (runLeft)
					dir = iDir - 30;
				else
					dir = iDir + 30;
			else
				if (runLeft)
					dir = iDir - 10;
				else
					dir = iDir + 10;
			
			minDir = 15;
		}
	}
	else 
		// no one is in way
		// calculate the preciseness for direction
		minDir = GetDirectionPreciseness(dist);
	
	// calculate if the agent has to turn and how much
//	deltaBallDir = pAgent->GetAngleTo(*m_pPerception->GetFutureBallPosition(1)) - dir;
//	actualTurnDir = dir + headAngle; 
	ballDir = pAgent->GetAngleToBody(*m_pPerception->GetFutureBallPosition(1));
	deltaBallDir = ballDir - dir;
	actualTurnDir = dir;

	// if we have to turn close to 180 deg, we turn in ball direction
	if (actualTurnDir < -170)
		if (ballDir < 0)
			actualTurnDir = -179;
		else
			actualTurnDir = 180;
	else if (actualTurnDir > 170)
		if (ballDir < 0)
			actualTurnDir = -179;
		else
			actualTurnDir = 180;


	if (bMindBall && ((deltaBallDir < -90 - viewAngle) || (deltaBallDir > 90 + viewAngle))
		&& (pAgent->GetNum() != GOALIE_ID))
		// players do not dash backward
		bMindBall = FALSE;

	if (bMindBall && ((deltaBallDir < -90 - viewAngle) || (deltaBallDir > 90 + viewAngle)))
	{
		// only the goalie dashes backward
		deltaBallDir = deltaBallDir - 180 - headAngle;
		// turn head to ball
		TurnNeck(deltaBallDir);

		// player must dash backwards to keep ball in mind
		actualTurnDir -= 180;
		if ((actualTurnDir < -minDir) || (actualTurnDir > minDir))
			// first turn into direction
			// turn body
			Turn(actualTurnDir);
		else
			// we already head the proper direction
			Dash(-iPower);
	}
	else
	{
		willTurn = (actualTurnDir < -minDir) || (actualTurnDir > minDir);

		if (bMindBall)
		{
			if (!checkSourrounding || (checkSourrounding && !CheckSourrounding()))
				// turn head into direction of ball
				TurnNeck(deltaBallDir - headAngle);
		}
		else
			// turn head into direction of dashing
			TurnNeck(-(int)headAngle);

		if (willTurn)
			// first turn into direction
			// turnbody
			Turn(actualTurnDir);
		else
			Dash(iPower);
	}
}

/*-----------------------------------------------------------------------------
Function: int TDoingAgent::GetDirectionPreciseness(double dBallDistance)
Date    : 22.07.99
-------------------------------------------------------------------------------
Parameter: distance of the ball
Returnval: angle the agent may head into the wrong direction when dashing
Description: to avoid too often turning, the agent needs not precisely dash into
				 the specified direction. This method returns the error in direction the
				 agent may have + / -
-----------------------------------------------------------------------------*/
int TDoingAgent::GetDirectionPreciseness(double dBallDistance)
{
	if (m_pPerception->GetThisPlayer()->GetNum() == GOALIE_ID)
	{
		// goalie does not need to dash precise near ball
		if (dBallDistance > 8.0)
			return 6;
		else if (dBallDistance < 1.8)
			// if the ball is so close we have to be precise again, since
			// we obviously had a catch fault
			return 50;
		else
			return (int) (asin(1.0 / (dBallDistance - 0.8)) * RAD2DEG) - 2;
	}
	else
	{
		// normal player
		if (dBallDistance > 8.0)
			return 7;
		else if (dBallDistance < 1.2)
			return 40;
		else
			return (int) (asin(1.0 / (dBallDistance + 0.3)) * RAD2DEG);
	}
}

/*-----------------------------------------------------------------------------
Function: BOOL TDoingAgent::CheckSourrounding()
Date    : 01.03.2000
-------------------------------------------------------------------------------
Parameter: none
Returnval: TRUE, if the agent turned its neck
Description: turns the head every 2 seconds from left to right to see the beauty
				 of the scene
-----------------------------------------------------------------------------*/
BOOL TDoingAgent::OldCheckSourrounding()
{
	//if the ball is far, we can look around
	double	dist = m_pPerception->GetObjectDistance(m_pPerception->GetBall());
	int		id = m_pPerception->GetThisPlayer()->GetNum();
	int		iSeeTime = m_pPerception->GetSeeTime();

	// before kick off, we do not look around
	if (iSeeTime == 0)
		return FALSE;

	if ((dist > 15.0) && (id != GOALIE_ID))
	{
		// from time to time we turn our head
		if ((iSeeTime % 40) == 0)
			TurnNeck(-180);
		else if ((iSeeTime % 20) == 0)
			TurnNeck(180);
		else
			return FALSE;

		SetWaitForNextSee();
		return TRUE;
	}
	return FALSE;
}

BOOL TDoingAgent::CheckSourrounding()
{
	//if the ball is far, we can look around
	double	dist = m_pPerception->GetObjectDistance(m_pPerception->GetBall());
	int		id = m_pPerception->GetThisPlayer()->GetNum();
	int		time = m_pPerception->GetSenseTime();
	int		nextSayTimeDelta = GetNextSayTime() - time;
	CString	msg;

	// before kick off, we do not look around
	if (time == 0)
		return FALSE;

	// only midfielders do
	msg.Format("CheckSourrounding nextSayTime:%d ballDist:%4.2f seeWideCount:%d", nextSayTimeDelta, dist, m_seeWideCount);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
	if ((dist > 20.0) && (id >MAX_DEFENDER) && (id <= PLAYERS_PER_TEAM))
	{
		if ((id <= MAX_MIDFIELDER) && (m_pPerception->GetObjectDistance(m_pPerception->GetOwnGoal()) < 40.0))
			// once midfielders hide opponents they have no time to look around
			return FALSE;

		// from time to time we turn our head (always before saying something)
		if (m_pPerception->GetTimeProceeding())
		{
			if (nextSayTimeDelta < 8)
			{
				// start to look around
				m_bWantWideView = TRUE;

				if (nextSayTimeDelta == 7)
				{
					// look to the left
					m_seeWideCount = 0;
					TurnNeck(-180);
					ChangeView(WIDE_VIEW, 1);
					m_pView->DisplayDebug("checkSour, time proc, turnneck left", KATEGORY_ACTION, DEBUG_LOWACTION);
					return TRUE;
				}
				
				if (m_seeWideCount == 1)
				{
					// look to the right
					TurnNeck(180);
					ChangeView(WIDE_VIEW, 1);
					m_pView->DisplayDebug("checkSour, time proc, turnneck right", KATEGORY_ACTION, DEBUG_LOWACTION);
					return TRUE;
				}
				
				if (m_seeWideCount == 2)
				{
					// now we should have seen all
					m_pPerception->KnowOffside();
					m_bWantWideView = FALSE;
					ChangeView(NORMAL_VIEW, 1);
					m_pView->DisplayDebug("checkSour, time proc, saw all", KATEGORY_ACTION, DEBUG_LOWACTION);
					m_seeWideCount = 0;
					return FALSE;
				}
				
				return TRUE;
			}
		}
		else
		{
			// the game is not running, so always look 180 degree
			m_bWantWideView = TRUE;
			ChangeView(WIDE_VIEW, 1);

			if (m_seeWideCount == 0)
			{
				// look to the left
				TurnNeck(-180);
				SetWaitForNextSee();
				m_pView->DisplayDebug("checkSour, time not, turn left", KATEGORY_ACTION, DEBUG_LOWACTION);
			}
			else if (m_seeWideCount == 1)
			{
				TurnNeck(180);
				SetWaitForNextSee();
				m_pView->DisplayDebug("checkSour, time not, turn right", KATEGORY_ACTION, DEBUG_LOWACTION);
			}
			else
			{
				// we have seen all so start again
				m_pPerception->KnowOffside();
				TurnNeck(-180);
				SetWaitForNextSee();
				m_pView->DisplayDebug("checkSour, time not, saw all", KATEGORY_ACTION, DEBUG_LOWACTION);
				m_seeWideCount = 0;
			}
		}
	}
	return FALSE;
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::ComplexKick(int iPower, int iDirection)
Date    : 15.07.98
-------------------------------------------------------------------------------
Parameter: power and relative direction in which to kick
Returnval: (void)
Description: avoids to kick onto the agent itself and takes the speed of the 
				 ball into account
-----------------------------------------------------------------------------*/
void TDoingAgent::ComplexKick(int iPower, int iDirection, bool bQueue)
{
	TDirection	ballDir = m_pPerception->GetObjectDirection(m_pPerception->GetBall());
	double		ballDist = m_pPerception->GetObjectDistance(m_pPerception->GetBall());
	TDirection	angle = ballDir - iDirection;
	int			newPower = iPower;
	TDirection	newDir = iDirection;
	TPlayer*	opponent = m_pPerception->GetOpponentAtMe();
	TDirection	opponentDir;
	CString		msg;
	int			resultPower = 0;
	int			resultDir = 0;
	int			iAbsX = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int			iAbsY = (int) m_pPerception->GetThisPlayer()->GetAbsY();
	double		resultFactor;

	if (m_pPerception->GetSide() != LEFT_TO_RIGHT)
	{
		iAbsX = -iAbsX;
		iAbsY = -iAbsY;
	}

	msg.Format("ComplexKick %d %d", iPower, iDirection);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_MIDACTION);

	if (GetIsBehind(ballDist, angle))
	{
		// ball is behind agent, so bring it to either side depending on opponents near
		if ((opponent != NULL) && (m_pPerception->GetObjectDistance(opponent) < 4.0))
		{
			opponentDir = m_pPerception->GetObjectDirection(opponent) - iDirection;
			if (opponentDir > 0)
				// opponent is right, so dribble ball to left
				DribbleClockwise(HARD_DRIBBLE);
			else
				// opponent is left, so dribble ball to right
				DribbleCounterclockwise(HARD_DRIBBLE);
			return;
		}
		else
		{
			// in own half always dribble away from own goal
			if (iAbsX < 0)
			{
				// in own half always dribble away from own goal
				if (iAbsY > 0)
					DribbleCounterclockwise(HARD_DRIBBLE);
				else
					DribbleClockwise(HARD_DRIBBLE);
			}
			else if (angle > 0)
				DribbleCounterclockwise(HARD_DRIBBLE);
			else
				DribbleClockwise(HARD_DRIBBLE);
			return;
		}
	}

	// take the speed of the ball into account
	AccountBallSpeed(newPower, newDir, resultPower, resultDir);
	
	// calculate the factor that is satisfying, in other half we want high result
	if (iAbsX < -30)
		resultFactor = 0.3;
	else if (iAbsX < 0)
		resultFactor = 0.6;
	else
		resultFactor = 0.8;

	if ((newPower < 5) && (iPower > 7))
	{
		// the ball already runs into the desired direction, so run to it
		// ???
		ComplexDashTo(*m_pPerception->GetFutureBallPosition(1), 80, 0);
	}
	else if ((newPower >= 100) && (iPower > 0) && ((double)resultPower / (double)iPower < resultFactor))
	{
		// the expected ball speed can only be reached to a certain degree (50%)
		StopBall();
		msg.Format("kick too weak: resultPower:%d wantedPower:%d", resultPower, iPower);
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
	}
	else
	{
		// if the player can not kick in desired direction, stop ball
		int			headAngle	= m_pPerception->GetHeadAngle();
		TDirection	delta		= iDirection + headAngle - resultDir;
		delta.Abs();
		if (delta > 2)
		{
			// the expected ball speed can only be reached to a certain degree (50%)
			StopBall();
			msg.Format("kick too weak: resultDir:%d wantedDir:%d", resultDir, iDirection + headAngle);
			m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		}
		else
			// normal kick
			Kick(newPower, newDir, bQueue);
	}
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::AccountBallSpeed(int& power, TDirection& dir)
Date    : 25.08.98
-------------------------------------------------------------------------------
Parameter: power of kicking
			  relative direction in which to kick
Returnval: (void)
Description: take the relative speed vector of the ball into account as well as 
				 the direction in which to kick
-----------------------------------------------------------------------------*/
void TDoingAgent::AccountBallSpeed(int& power, TDirection& dir, int& resultPower, int& resultDir)
{
	TBall*			pBall = m_pPerception->GetBall();
	TThisPlayer*	pAgent = m_pPerception->GetThisPlayer();
	double			dBallDist = pBall->GetDistance() - m_pPerception->GetPlayerSize() - BALL_SIZE;
//	double			dDeltaDir = (double) pBall->GetDirection() / 180.0;
	double			dDeltaDir = (double) pAgent->GetAngleToBody(pBall->GetPosition()) / 180.0;
//	TVector			tBallSpeed = pBall->GetRelSpeed(pAgent->GetPosition(), pAgent->GetAbsDir());
	TVector			absBallSpeed = pBall->GetAbsSpeed();
	TVector			tBallSpeed = pBall->GetRelSpeed(pAgent->GetPosition(), pAgent->GetAbsBodyDir());
	double			dBallSpeed = tBallSpeed.GetLength();
	TVector			tWantToKick;
	TVector			resultingBallSpeed;
	double			dKickLen;
	double			dFactor;
	CString			tmp;
	CString			msg;
	TDirection		bodyDir;

	if (dDeltaDir < 0.0)
		dDeltaDir = -dDeltaDir;
	if (dDeltaDir > 1.0)
		// should not happen
		dDeltaDir = 1.0;

	if (dBallDist > m_pPerception->GetKickableMargin())
		// should not happen, since kick should only be executed if ball is kickable
		dBallDist = m_pPerception->GetKickableMargin();
	else if (dBallDist < 0.0)
		dBallDist = 0.0;

	// take head angle into account
	bodyDir = dir + m_pPerception->GetHeadAngle();
	msg.Format("Account Ball Speed: old power: %d old dir: %d old dirbody: %d bsx:%4.2f bsy:%4.2f babsx:%4.2f babsy:%4.2f"
		, power, (int)dir, (int) bodyDir, tBallSpeed.GetX(), tBallSpeed.GetY(), absBallSpeed.GetX(), absBallSpeed.GetY());

	// calculate the wanted final ball speed
	dFactor = 1.0 - 0.25 * dDeltaDir - 0.25 * dBallDist / m_pPerception->GetKickableMargin();
	tWantToKick.SetVectorByDirection(power * KICK_POWER_RATE /* * dFactor*/, (int) bodyDir);
	dKickLen = tWantToKick.GetLength();

	if (power >= 100)
	{
		// a full power kick does not slow down the ball if it already goes into correct dir
		// so we assume that the want to kick is ballspeed + 100 power

//		tWantToKick *= ((BALL_SPEED_MAX + 100.0 * KICK_POWER_RATE) / dKickLen);
//		tWantToKick -= tBallSpeed;

		tWantToKick = CalculateKick(bodyDir, tBallSpeed.GetX(), tBallSpeed.GetY(), power * KICK_POWER_RATE * dFactor);
	}
	else
		tWantToKick -= tBallSpeed; 

	dir = tWantToKick.GetDirection();
	power = (int) (tWantToKick.GetLength() / (KICK_POWER_RATE * dFactor));

	if (power > 100)
		power = 100;

	// check if the resulting ball speed is somewhere near what is expected
	resultingBallSpeed = tWantToKick + tBallSpeed;
	resultPower = (int) (resultingBallSpeed.GetLength() / KICK_POWER_RATE);	
	resultDir = resultingBallSpeed.GetDirection();

	tmp.Format("  new power: %d new dir: %d resultPower:%d, resultDir:%d", power, (int)dir, resultPower, resultDir);
	m_pView->DisplayDebug(msg + tmp, KATEGORY_ACTION, DEBUG_LOWACTION);

	#ifdef SHOW_DOING_VAR
		TRACE3("account ball speed: ball speed: %4.2f, balldir: %d, ballpower: %4.2f\n"
					, pBall->GetRelSpeed(pAgent->GetPosition(), pAgent->GetAbsDir()).GetLength()
					, (int)tBallSpeed.GetDirection()
					, tBallSpeed.GetLength());

		TRACE2("account ball speed after: power: %d, dir: %d\n", power, (int)dir);
	#endif
}

/*-----------------------------------------------------------------------------
Function: BOOL TDoingAgent::DribbleCounterclockwise(int power)
Date    : 26.08.98
-------------------------------------------------------------------------------
Parameter: power with which to kick
Returnval: true, if a kick has been executed
Description: dribbles ball in counterclockwise direction
			 the higher the power, the likelier the ball is lost
-----------------------------------------------------------------------------*/
BOOL TDoingAgent::DribbleCounterclockwise(int power)
{
	double		size = m_pPerception->GetPlayerSize() + BALL_SIZE + 0.15;
	TBall*		pBall = m_pPerception->GetBall();
	TThisPlayer* pAgent = m_pPerception->GetThisPlayer();
//	TDirection	ballDir = m_pPerception->GetObjectDirection(pBall);
	TDirection	ballDir = pAgent->GetAngleToBody(pBall->GetPosition());
	double		ballDist = m_pPerception->GetObjectDistance(pBall);
	int			resultPower = 0;
	int			resultDir = 0;
	TDirection	dir;

	m_pView->DisplayDebug("DribbleCounterclockwise", KATEGORY_ACTION, DEBUG_MIDACTION);

	// avoid kicking into outside if kick in
	if (m_pPerception->GetPlaymodeKickInOwn())
	{
		dir = pAgent->GetAngleToBody(m_pPerception->GetOther11mPoint()->GetPosition());
	}
	else if (ballDist < size)
	{
		// the ball is very close
		// !!! check if the ball has collided with player and has now zero speed
		dir = ballDir - 90;
		power -= (int) (power * m_geneticCode->m_parameters[CLOSE_DRIBBLE_POWER_INDEX]);
	}
	else
	{
		// kick close aside the player
		dir = ballDir - 180 + (int)(asin(size / ballDist) * RAD2DEG)
				+ (int) m_geneticCode->m_parameters[DRIBBLE_ANGLE_INDEX];
	}
	
	AccountBallSpeed(power, dir, resultPower, resultDir);
	
	if (power > 5)
	{
		Kick(power, dir);
		return TRUE;
	}
	else
		return FALSE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TDoingAgent::DribbleClockwise(int power)
Date    : 26.08.98
-------------------------------------------------------------------------------
Parameter: power with which to kick
Returnval: true, if a kick has been executed
Description: dribbles ball clockwise around the player
-----------------------------------------------------------------------------*/
BOOL TDoingAgent::DribbleClockwise(int power)
{
	double		size = m_pPerception->GetPlayerSize() + BALL_SIZE + 0.15;
	TBall*		pBall = m_pPerception->GetBall();
	TThisPlayer* pAgent = m_pPerception->GetThisPlayer();
//	TDirection	ballDir = m_pPerception->GetObjectDirection(pBall);
	TDirection	ballDir = pAgent->GetAngleToBody(pBall->GetPosition());
	double		ballDist = m_pPerception->GetObjectDistance(pBall);
	TDirection	dir;
	int			resultPower = 0;
	int			resultDir = 0;

	m_pView->DisplayDebug("DribbleClockwise", KATEGORY_ACTION, DEBUG_MIDACTION);

	// avoid kicking into outside if kick in
	if (m_pPerception->GetPlaymodeKickInOwn())
	{
		dir = pAgent->GetAngleToBody(m_pPerception->GetOther11mPoint()->GetPosition());
	}
	else if (ballDist < size)
	{
		// the ball is very close
		dir = ballDir + 90;
		power -= (int) (power * m_geneticCode->m_parameters[CLOSE_DRIBBLE_POWER_INDEX]);
	}
	else
	{
		// kick close aside the player
		dir = ballDir + 180 - (int)(asin(size / ballDist) * RAD2DEG)
				- (int) m_geneticCode->m_parameters[DRIBBLE_ANGLE_INDEX];
	}
	AccountBallSpeed(power, dir, resultPower, resultDir);
	if (power > 5)
	{
		Kick(power, dir);
		return TRUE;
	}
	else
		return FALSE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TDoingAgent::MoveBallToPoint(int iDirection, int iOrientation)
Date    : 10.02.2000
-------------------------------------------------------------------------------
Parameter: relative direction to the head of the agent, the ball should be placed to
			  flag, indicating if the ball should be dribbled clockwise (-1),
			  counterclockwise (1) or if the caller does not mind (0)
Returnval: true, if a kick has been executed
Description: moves the ball to the desired direction and stops it there
-----------------------------------------------------------------------------*/
BOOL TDoingAgent::MoveBallToPoint(int iDirection, int iOrientation)
{
	TDirection	ballDir = m_pPerception->GetObjectDirection(m_pPerception->GetBall());
	TDirection	deltaDir = ballDir - iDirection;
	int			resultPower = 0;
	int			resultDir = 0;
	CString		msg;

	msg.Format("MoveBallToPoint dir: %d orientation: %d", iDirection, iOrientation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_MIDACTION);

	// look if we can kick directly to the desired point
	if (deltaDir > 90)
	{
		if (iOrientation == CLOCKWISE)
			// we dribble the long way, because caller wants this
			return DribbleClockwise(SECURE_DRIBBLE);
		else
			return DribbleCounterclockwise(SECURE_DRIBBLE);
	}
	else if (deltaDir < -90)
	{
		if (iOrientation == COUNTERCLOCKWISE)
			// we dribble the long way, because caller wants this
			return DribbleCounterclockwise(SECURE_DRIBBLE);
		else
			return DribbleClockwise(SECURE_DRIBBLE);
	}
	else
	{
		// the ball may be at the desired point
		// calculate the vector from ball to desired point
		TThisPlayer* pAgent = m_pPerception->GetThisPlayer();
		TDirection	absGoalDir = pAgent->GetAbsDir() + iDirection;
		double		dDistance = m_pPerception->GetPlayerSize() + BALL_SIZE + m_pPerception->GetKickableMargin() - 0.4;
		TVector		agentPos = pAgent->GetPosition();
		TVector		goalPoint = m_pPerception->GetAbsCoordinates(agentPos, dDistance, absGoalDir);
		TVector		ballPos = m_pPerception->GetBall()->GetPosition();
		TVector		deltaPos = agentPos + (goalPoint - ballPos);
		TDirection	dir = pAgent->GetAngleToBody(deltaPos);
		// we multiplied by 3 to be cautious
		int			power = (int) ((goalPoint - ballPos).GetLength() / (3.0 * KICK_POWER_RATE));

		if (power > 100)
			power = 100;

		AccountBallSpeed(power, dir, resultPower, resultDir);
		if (power > 5)
		{
			Kick(power, dir);	
			return TRUE;
		}
		else
			return FALSE;
	}
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::SurroundOpponent(int dir)
Date    : 22.02.2000
-------------------------------------------------------------------------------
Parameter: direction, which is desired to run to
Returnval: (void)
Description: tries to dribble around an opponent
-----------------------------------------------------------------------------*/
void TDoingAgent::SurroundOpponent(int direction)
{
	CString		msg;
	int			iCount = 6;
	double		optimalDist = 0.0;
	TDirection	optimalDir = direction;
	TDirection	dir;
	double		dist;

	msg.Format("SurroundOpponent dir: %d", direction);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_MIDACTION);

	// we try to run past the opponent
	// find the direction where most space is
	dir = direction - (5 * iCount);
	for (int i = 0; i <= iCount; i++)
	{
		dist = m_pPerception->GetAccessDistance(dir);
		if (dist > optimalDist)
		{
			optimalDir = dir;
			optimalDist = dist;
		}
		dir = dir + 10;
	}

	ComplexDashAndKick(100, optimalDist, optimalDir);
}

/*-----------------------------------------------------------------------------
Function: BOOL TDoingAgent::GetIsBehind(double dist, int dir)
Date    : 23.07.99
-------------------------------------------------------------------------------
Parameter: distance of the point / object
			  relative delta direction between ball and wanted kick direction
Returnval: (void)
Description: checks if the ball is behind the agent, so that it can not kick
				 directly into kick direction
-----------------------------------------------------------------------------*/
BOOL TDoingAgent::GetIsBehind(double dist, int angle)
{
/*	// better use TArea
	if ((dist < 0.4) && ((angle < -100) || (angle > 100)))
		return TRUE;
	if ((dist < 0.5) && ((angle < -145) || (angle > 145)))
		return TRUE;
	if ((dist < 0.7) && ((angle < -155) || (angle > 155)))
		return TRUE;
	if ((dist < 1.1) && ((angle < -160) || (angle > 160)))
		return TRUE;
*/
	if (angle < -90)
		angle = 180 + angle;
	else if (angle > 90)
		angle = 180 - angle;
	else
		return FALSE;

	if (dist * sin(DEG2RAD * angle) < PLAYER_SIZE / 2.0)
		return TRUE;
	else
		return FALSE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TDoingAgent::GetIsInFront(double dist, int angle)
Date    : 22.08.00
-------------------------------------------------------------------------------
Parameter: distance of the point / object
			  relative delta direction between ball and wanted kick direction
Returnval: (void)
Description: checks if the ball is in front of the agent (used to avoid running into
				 the ball, so distance is less
-----------------------------------------------------------------------------*/
BOOL TDoingAgent::GetIsInFront(double dist, int angle)
{
	// better use TArea
	if ((dist < 0.45) && (angle > -90) && (angle < 90))
		return TRUE;
	if ((dist < 0.55) && (angle > -40) && (angle < 40))
		return TRUE;
	if ((dist < 0.7) && (angle > -32) && (angle < 32))
		return TRUE;
	return FALSE;
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::HardKick(int iDirection)
Date    : 16.07.99
-------------------------------------------------------------------------------
Parameter: relative direction into which to kick
Returnval: (void)
Description: dribbles the ball into proper position for a double kick
			 assumes that the player has got enough time !
-----------------------------------------------------------------------------*/
void TDoingAgent::HardKick(int iDirection)
{
	TBall*		pBall = m_pPerception->GetBall();
	TThisPlayer* pAgent = m_pPerception->GetThisPlayer();
	TDirection	ballDir = m_pPerception->GetObjectDirection(pBall);
	double		ballDist = m_pPerception->GetObjectDistance(pBall);
	TDirection	angle = ballDir - iDirection;
	int			iHeadAngle = m_pPerception->GetHeadAngle();
	int			iAgentX;
	double		dBallSpeed = pBall->GetAbsSpeed().GetLength();
	int			iBallSpeedDir;
	double		dRelBallSpeed;
	TVector		relBallSpeed = pBall->GetRelSpeed(pAgent->GetPosition(), pAgent->GetAbsBodyDir());
	TDirection	dirToBody = iDirection + iHeadAngle;
	TDirection	ballSpeedDelta;
	BOOL		bKickTwice = FALSE;
	CString		msg;

	// check if we could kick the ball twice
	if (!GetIsBehind(ballDist, angle) && ((angle < -100) || (angle > 100)))
		bKickTwice = TRUE;
		
	iBallSpeedDir = relBallSpeed.GetDirection();
	dRelBallSpeed = relBallSpeed.GetLength();
	ballSpeedDelta = iBallSpeedDir - iDirection;

	msg.Format("HardKick dir:%d relBallDir:%d relBallSpeed:%4.2f delta:%d", 
				iDirection, iBallSpeedDir, dRelBallSpeed, (int) ballSpeedDelta);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_MIDACTION);


	double minBallSpeed = m_geneticCode->m_parameters[MIN_BALL_SPEED_INDEX];
	int minTurnDir = (int) m_geneticCode->m_parameters[MIN_TURN_DIR_INDEX];

	if (m_pPerception->GetPlayMode() != PLAY_ON) 
	{
		// if freekickmode, we run close to the ball 
		if (ballDist > PLAYER_SIZE + 0.3)
		{
			ComplexDashTo(pBall, 20, 0);
			return;
		}
		// and turn into kickdirection
		if ((dirToBody < -10) || (dirToBody >10))
		{
			// turn into kick direction
			Turn(dirToBody);
			return;
		}
		// and wait for visual
		if (m_pPerception->GetSeeTime() < m_pPerception->GetSenseTime())
		{
			return;
		}
	}

	// ??? could also test for ball is close
	if (dRelBallSpeed < minBallSpeed)
	{
		// kick the ball
		ComplexKick(100, iDirection);

		if (bKickTwice)
			Kick(100, iDirection, TRUE);

		msg.Format("hardkick kick: angle %d distance %4.2f ballspeed %4.2f kick2:%d"
			, (int) angle, ballDist, dRelBallSpeed), bKickTwice;
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		return;
	}
	
	iAgentX = (int) pAgent->GetAbsX();
	if (m_pPerception->GetSide() == RIGHT_TO_LEFT)
		iAgentX = -iAgentX;

	if ((iAgentX < -20) || (ballSpeedDelta > -50) || (ballSpeedDelta < 50))
	{
		// no hardkick in own half or if the ball is already proper direction
		ComplexKick(100, iDirection);
		if (bKickTwice)
			Kick(100, iDirection, TRUE);

		msg.Format("No hardkick, own half or ball direction ok: %4.2f", dBallSpeed);
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		return;
	}

	// ???
	if (m_pPerception->GetOpponentNearBall(2.5))
	{
		// we have no time because opponent is near
		ComplexKick(100, iDirection);
		if (bKickTwice)
			Kick(100, iDirection, TRUE);

		msg.Format("No hardkick, opponent too close1");
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		return;
	}

	if (minTurnDir < 0)
		minTurnDir = 0;
	if (WillHaveBallNextStep())
	{
		// we have time to do turn
		if ((dirToBody < -minTurnDir) || (dirToBody > minTurnDir))
		{
			// turn into kick direction
			Turn(dirToBody);
			msg.Format("hardkick, first turn: %d", (int) dirToBody);
			m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
			return;
		}
	}

	if (GetIsBehind(ballDist, angle))
	{
		// ball is behind, so put it to the side
		ComplexKick(100, iDirection);
		return;
	}

	// in any other case we stop the ball
	StopBall();
}

/*-------------------------------------------------------------------------------
Parameter: none
Returnval: true if the ball will be kickable next step
Description: checks if the ball will be kickable next step
-----------------------------------------------------------------------------*/
BOOL TDoingAgent::WillHaveBallNextStep()
{
	TVector		tFutureAgent;
	TVector		tFutureBall = *m_pPerception->GetFutureBallPosition(1);
	double		dMaxKick = m_pPerception->GetPlayerSize() + BALL_SIZE + m_pPerception->GetKickableMargin() - 0.2;
	double		dFutureDistance;

	m_pPerception->GetThisPlayer()->GetFuturePosition(1, tFutureAgent);
	dFutureDistance = tFutureBall.GetDistanceTo(tFutureAgent);
	if (dFutureDistance < dMaxKick)
		return TRUE;
	else
		return FALSE;
}

/*-----------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: stops the ball by kicking it against the agent
-----------------------------------------------------------------------------*/
void TDoingAgent::StopBall()
{
	TBall*		pBall			= m_pPerception->GetBall();
/*	TDirection	ballDir			= m_pPerception->GetObjectDirection(pBall);
	double		ballDist		= m_pPerception->GetObjectDistance(pBall);
	double		stopKickPower	= m_geneticCode->m_parameters[STOP_KICK_POWER_INDEX];
*/	int			resultPower		= 0;
	int			resultDir		= 0;
	int			power;
	CString		msg;
	TDirection	dir	;
	
	dir = 0;
	power = 0;
	AccountBallSpeed(power, dir, resultPower, resultDir);
	Kick(power, dir);

	msg.Format("stop ball: power:%d dir:%d ballspeed:%4.2f", 
				power, (int) dir, pBall->GetAbsSpeed().GetLength());
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
	return;

/*	// stop the ball by kicking it against the agent
	dir -= 180;
	power = (int) (ballDist * stopKickPower);
	AccountBallSpeed(power, dir, resultPower, resultDir);
	Kick(power, dir);

	msg.Format("stop ball: distance %4.2f dir %d ballspeed %4.2f", 
				ballDist, (int) ballDir, pBall->GetAbsSpeed().GetLength());
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
	return;
*/
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::HardKick(int iDirection)
Date    : 16.07.99
-------------------------------------------------------------------------------
Parameter: relative direction into which to kick
Returnval: (void)
Description: dribbles the ball into proper position for a double kick
			 assumes that the player has got enough time !
-----------------------------------------------------------------------------*/
void TDoingAgent::OldHardKick(int iDirection)
{
	TBall*		pBall = m_pPerception->GetBall();
	TThisPlayer* pAgent = m_pPerception->GetThisPlayer();
	TDirection	ballDir = m_pPerception->GetObjectDirection(pBall);
	double		ballDist = m_pPerception->GetObjectDistance(pBall);
	TDirection	angle = ballDir - iDirection;
	double		dMaxKick = m_pPerception->GetPlayerSize() + BALL_SIZE + m_pPerception->GetKickableMargin() - 0.2;
	TVector		tFutureAgent;
	TVector		tFutureBall = *m_pPerception->GetFutureBallPosition(1);
	double		dFutureDistance;
	TDirection	iFutureDirection;
	int			iHeadAngle = m_pPerception->GetHeadAngle();
	int			iAgentX;
	CString		msg;
	double		dBallSpeed = pBall->GetAbsSpeed().GetLength();
	int			iBallSpeedDir;
	double		dRelBallSpeed;
	TVector		relBallSpeed = pBall->GetRelSpeed(pAgent->GetPosition(), pAgent->GetAbsBodyDir());

	int			minKickAngle;	
	int			firstKickPower;
	int			secondKickPower;
	int			secureDribblePower;

	msg.Format("HardKick dir: %d", iDirection);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_MIDACTION);

	iAgentX = (int) pAgent->GetAbsX();
	if (m_pPerception->GetSide() == RIGHT_TO_LEFT)
		iAgentX = -iAgentX;

	iBallSpeedDir = relBallSpeed.GetDirection();
	dRelBallSpeed = relBallSpeed.GetLength();
	
	if ((iAgentX < -5) || ((dBallSpeed > 0.5)
		 && ((iBallSpeedDir < -20) || (iBallSpeedDir > 20))))
	{
		// no hardkick in own half or if the ball is fast but not running away
		ComplexKick(100, iDirection);
		msg.Format("No hardkick, own half or ball too fast: %4.2f", dBallSpeed);
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		return;
	}

	if (m_pPerception->GetOpponentNearBall(2.5))
	{
		// we have no time because opponent is near
		ComplexKick(100, iDirection);
		msg.Format("No hardkick, opponent too close1");
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		return;
	}

	// if the ball will stay close, we can first look if it is good to turn into kick
	// kick direction
	TDirection dirToBody = iDirection + iHeadAngle;
	pAgent->GetFuturePosition(1, tFutureAgent);
	dFutureDistance = tFutureBall.GetDistanceTo(tFutureAgent);

	if (dFutureDistance < dMaxKick)
	{
		// we have time to do turn
		if ((dirToBody < -20) || (dirToBody > 20))
		{
			// turn into kick direction
			Turn(dirToBody);
			msg.Format("hardkick, first turn: %d", (int) dirToBody);
			m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
			return;
		}
	}

	if (GetIsBehind(ballDist, angle))
	{
		// ball is behind, so put it to the side
		ComplexKick(100, iDirection);
		return;
	}

	minKickAngle = (int) MIN_KICK_ANGLE;
	firstKickPower = (int) FIRST_KICK_POWER;
	secondKickPower = (int) SECOND_KICK_POWER;
	secureDribblePower = (int) SECURE_DRIBBLE;
/*
	minKickAngle = (int) m_geneticCode->m_parameters[MIN_KICK_ANGLE_INDEX];
	firstKickPower = (int) m_geneticCode->m_parameters[FIRST_KICK_POWER_INDEX];
	secondKickPower = (int) m_geneticCode->m_parameters[SECOND_KICK_POWER_INDEX];
	secureDribblePower = (int) m_geneticCode->m_parameters[SECURE_DRIBBLE_INDEX];
*/
	if ((angle < -minKickAngle) || (angle > minKickAngle))
	{
		// ok, we kick hard
		int theAngle = angle;
		if (theAngle < 0)
			theAngle = - theAngle;

		secondKickPower = theAngle - minKickAngle + secondKickPower;
		if (secondKickPower > 100)
			secondKickPower = 100;
		if (secondKickPower < 10)
			secondKickPower = 10;


		// do the kick
//		if (dBallSpeed > 0.2)
//		{
			// we do three kicks first stop the ball, 2 and 3 accelerate
//			ComplexKick(firstKickPower, dirToBody);
//			Kick(secondKickPower, dirToBody, TRUE);
//		}
//		else
			// we do two kicks
//			Kick(secondKickPower, dirToBody);

		ComplexKick(secondKickPower, dirToBody);
		Kick(100, dirToBody, TRUE);

		msg.Format("hardkick done3: angle %d distance %4.2f ballspeed %4.2f", (int) angle, ballDist, dBallSpeed);
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		return;
	}

	// the ball is in front, so dash to it if time
	msg.Format("hardkick: futureDistance: %4.2f ball distance: %4.2f balldir: %d rel speed dir: %d ball speed: %4.2f"
					, dFutureDistance, ballDist, (int) ballDir, iBallSpeedDir, dRelBallSpeed);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
	
	if (!GetIsInFront(ballDist, dirToBody)
		&& (((iBallSpeedDir > -25) && (iBallSpeedDir < 25)) || (dRelBallSpeed < 0.2)))
	{
		// we can do a dash (we previously turned to ball dir)
		Dash(100);
		return;
	}
	
	// check if there is enough time for dribbling the ball
	if (m_pPerception->GetOpponentNearBall(6.0))
	{
		// we have no time because opponent is near
		ComplexKick(100, iDirection);
		msg.Format("No hardkick, opponent too close2");
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		return;
	}

	if (angle < 0)
		DribbleCounterclockwise(secureDribblePower);
	else
		DribbleClockwise(secureDribblePower);
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::ComplexDashAndKick(int iPower, double dist, int iDirection)
Date    : 15.07.98
-------------------------------------------------------------------------------
Parameter: power, with which to dash
			  distance that is secure to kick the ball
			  direction in which to dash
Returnval: (void)
Description: handles the dribbling of the ball. It is intentionally slower than
				 dashing without ball.
				 does not handle to dribble around a player by now
-----------------------------------------------------------------------------*/
void TDoingAgent::ComplexDashAndKick(int iPower, double dist, int iDirection)
{
	// direction is relative
	// player runs with the ball (dribling)
	TBall*	pBall = m_pPerception->GetBall();
	double	dMaxKickDistance = m_pPerception->GetPlayerSize() + m_pPerception->GetKickableMargin(); // + BALL_SIZE
	int		ballDir = m_pPerception->GetObjectDirection(pBall);
	double	ballDist = m_pPerception->GetObjectDistance(pBall);
	int		headAngle = m_pPerception->GetHeadAngle();
	CString	msg;


	msg.Format("ComplexDashAndKick pow:%d dist:%4.2f dir:%d", iPower, dist, iDirection);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_MIDACTION);

	//!!! if ball will be kickable if we dash then dash, but be aware of collisions

	if (ballDist > dMaxKickDistance)
	{
/*
		// will the ball be kickable next step
		TVector*	futureBall = m_pPerception->GetFutureBallPosition(1);
		double	futureBallDist = m_pPerception->GetThisPlayer()->GetDistanceTo(*futureBall);

		if (ballDist > dMaxKickDistance)
		{
			// run to the side of the ball

			if (ballDir - iDirection < 0)
				// run to the right side of the ball
				iDirection = ballDir + 10;
			else
				// run to the left side of the ball
				iDirection = ballDir - 10;

			ComplexDash(ballDist, iDirection, iPower);
		}
		else
		{
			// the ball will be kickable next cycle, so wait
			ComplexDash(0.0, iDirection, 0);			
		}
*/
		DoRunToBall(1.0);
	}
	else
	{
		// if we are not running into proper direction, we first turn and then kick
//		int			bodyDir = m_pPerception->GetThisPlayer()->GetAbsBodyDir();
		TDirection	delta	= iDirection + headAngle;
		double		ballSpeed = m_pPerception->GetBall()->GetAbsSpeed().GetLength();
		if ((delta > 20) || (delta < -20))
		{
			// check if we will have the ball next time, or if we should stop it
			if ((ballSpeed < 0.5) && (WillHaveBallNextStep()))
			{
				// turn into direction
				Turn(delta);
				TurnNeck(-headAngle);
				msg.Format("dashAndKick first turn dir:%d bspeed:%4.2f", iDirection, ballSpeed);
				m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
				return;
			}
			else if (ballSpeed > 0.2)
			{
				StopBall();
				msg.Format("dashAndKick stop ball dir:%d bspeed:%4.2f", iDirection, ballSpeed);
				m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
				return;
			}
		}
		
		// kick the ball
		int		iPower;
		// check if the player has got enough stamina to dribble as far as possible
		// is already done by TAccessMap
		iPower = (int) (GetLinearFuzzyValue(dist, 0.0, 40.0) * 1.2 / KICK_POWER_RATE) + 18; 
		ComplexKick(iPower, iDirection);
	}
}

/*-----------------------------------------------------------------------------
Function: BOOL TDoingAgent::ComplexDashTo(TVector position, int power, int dist)
Date    : 15.07.98
-------------------------------------------------------------------------------
Parameter: absolute field coordinates, where to run to
			  power, whith which to run (default 100)
			  radius of the region which is to reach
			  flag indicating, whether player should look at ball while running

Returnval: TRUE, if the coordinates already have been reached
Description: runs to the specified absolute place in the field. If the place 
				 has already been reached, no dash is done, but the function returns
				 TRUE, otherwise, the player dashes and returns FALSE
-----------------------------------------------------------------------------*/
BOOL TDoingAgent::ComplexDashTo(TVector& position, int power, int dist, BOOL mindBall, BOOL checkSourrounding)
{
	TDirection		direction;
	double			distance;
	TThisPlayer*	pAgent = m_pPerception->GetThisPlayer();

	// are we at the specified point ?
	if (m_pPerception->GetIsAt((int) position.GetX(), (int) position.GetY(), dist))
		return TRUE;

	// we are not there so where do we have to go
	
	distance = pAgent->GetDistanceTo(position);
	direction = pAgent->GetAngleToBody(position);

	// check if opponent team has freekick and we may not approach ball
	if (IsFreekickOther())
	{
		double distPointBall = m_pPerception->GetBall()->GetDistanceTo(position);
		double distMeBall = pAgent->GetDistanceTo(m_pPerception->GetBall());

		if ((distPointBall < 10.0) && (distMeBall < 10.0))
		{
			// point is inside not allowed area and I am almost there
			LookToBall();
			return TRUE;
		}
		else if (distMeBall < 13.0)
		{
			TDirection pointDir	= pAgent->GetAngleToBody(position);
			TDirection deltaDir = direction - pointDir;

			if (distMeBall < 10.0)
				if (deltaDir < 0)
					direction -= 90;
				else 
					direction += 90;
			else 
				if (deltaDir < 0)
					direction -= 70;
				else 
					direction += 70;
		}
	}

	ComplexDash(distance, direction, power, mindBall, checkSourrounding);

	return FALSE;
}

BOOL TDoingAgent::ComplexDashTo(TVisibleObject* pObj, int power, int dist, BOOL mindBall, BOOL checkSourrounding)
{
	return ComplexDashTo(pObj->GetPosition(), power, dist, mindBall, checkSourrounding);
}

BOOL TDoingAgent::IsFreekickOther()
{
	int	iPlayMode = m_pPerception->GetPlayMode();
	
	switch (iPlayMode)
	{
		case KICK_OFF_L:
		case KICK_IN_L:
		case GOAL_KICK_L:
		case FREE_KICK_L:
		case CORNER_KICK_L:
			// if this is a right team player, and is on position, it can wait
			// until game is play on
			if (GetSide() != LEFT_TO_RIGHT)
				return TRUE;
			break;

		case KICK_OFF_R:
		case KICK_IN_R:
		case GOAL_KICK_R:
		case FREE_KICK_R:
		case CORNER_KICK_R:
			// if this is a right team player, and is on position, it can wait
			// until game is play on
			// may be should be not dealt with as reflex
			if (GetSide() == LEFT_TO_RIGHT)
				return TRUE;
			break;
	}
	return FALSE;
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::ComplexKickTo(double x, double y, BOOL bStopThere)
Date    : 15.07.98
-------------------------------------------------------------------------------
Parameter: absolute x and y coordinates of the field
			  flag indicating, whether the ball should stop there (TRUE) or just be
			  kicked in that direction
Returnval: (void)
Description: kicks to the specified coordinates. If bStopThere is true, the kick
				 is softer, so that the ball almost stops at the specified position
-----------------------------------------------------------------------------*/
void TDoingAgent::ComplexKickTo(TVector& pos, BOOL bStopThere)
{
	int		power;
	int		direction = m_pPerception->GetThisPlayer()->GetAngleTo(pos);
	double	distance = m_pPerception->GetThisPlayer()->GetDistanceTo(pos);

	if (!bStopThere)
		power = 50 + (int) (GetLinearFuzzyValue(distance, 5.0, 20.0) * 50.0);
	else
		// the ball should almost stop there
		power = 10 + (int) (GetLinearFuzzyValue(distance, 5.0, 40.0) * 90.0);

	ComplexKick(power, direction);
}

void TDoingAgent::ComplexKickTo(TVisibleObject* pObj, BOOL bStopThere)
{
	ComplexKickTo(pObj->GetPosition(), bStopThere);
}


/*-----------------------------------------------------------------------------
Function: int TDoingAgent::GetPosY()
Date    : 11.08.00
-------------------------------------------------------------------------------
Parameter: none
Returnval: the y coordinate the player should keep
Description: calculates the y coordinate that should be the new home position y
-----------------------------------------------------------------------------*/
int TDoingAgent::GetPosY()
{
//	int	yBall		= (int) m_pPerception->GetBall()->GetAbsY();
//	int	xBall		= (int) m_pPerception->GetBall()->GetAbsX();
	int	yBall		= (int) m_pPerception->GetFutureBallPosition(15)->GetY();
	int	xBall		= (int) m_pPerception->GetFutureBallPosition(15)->GetX();
	int	number		= m_pPerception->GetThisPlayer()->GetNum();
	int	border1		= 20;
	int	border2		= 10;

	// if we have a goalkick, midfielders move
	if (m_pPerception->GetPlaymodeFreekickOwn() && 
		(m_pPerception->GetThisPlayer()->GetNum() <= MAX_MIDFIELDER))
	{
		if (GetSide() == RIGHT_TO_LEFT)
		{
			if (xBall > 40)
			{
				if (yBall < 0)
				{
//					if (number == MAX_MIDFIELDER)
						// the right midfielder moves to the outer line
						return -m_iPosKickOffLeft.y - 10;
				}
				else
				{
//					if (number == MAX_DEFENDER + 1)
						// the left midfielder moves to the outer line
						return -m_iPosKickOffLeft.y + 10;
				}
			}
		}
		else
		{
			if (xBall < -40)
			{
				if (yBall > 0)
				{
//					if (number == MAX_MIDFIELDER)
						// the right midfielder moves to the outer line
						return m_iPosKickOffLeft.y + 10;
				}
				else
				{
//					if (number == MAX_DEFENDER + 1)
						// the left midfielder moves to the outer line
						return m_iPosKickOffLeft.y - 10;
				}
			}
		}
	}


	if (GetSide() == RIGHT_TO_LEFT)
	{
		if (m_pPerception->GetThisPlayer()->GetNum() > MAX_MIDFIELDER)
			// offenders do not move
			return -m_iPosKickOffLeft.y;
	
		if (-yBall > border1)
			return -m_iPosKickOffLeft.y - 10;
		if (-yBall < -border1)
			return -m_iPosKickOffLeft.y + 10;
		if (-yBall > border2)
			return -m_iPosKickOffLeft.y - 5;
		if (-yBall < -border2)
			return -m_iPosKickOffLeft.y + 5;
		else
			return -m_iPosKickOffLeft.y;
	}
	else
	{
		if (m_pPerception->GetThisPlayer()->GetNum() > MAX_MIDFIELDER)
			// offenders do not move
			return m_iPosKickOffLeft.y;
		if (yBall > border1)
			return m_iPosKickOffLeft.y + 10;
		if (yBall < -border1)
			return m_iPosKickOffLeft.y - 10;
		if (yBall > border2)
			return m_iPosKickOffLeft.y + 5;
		if (yBall < -border2)
			return m_iPosKickOffLeft.y - 5;
		else
			return m_iPosKickOffLeft.y;
	}
}

/*-----------------------------------------------------------------------------
Function: TVector TDoingAgent::GetIntersectionPoint()
Date    : 21.08.00
-------------------------------------------------------------------------------
Parameter: none
Returnval: the coordinates where the player should intersect the ball
Description: calculates the coordinates where the player should intersect the ball
-----------------------------------------------------------------------------*/
TVector TDoingAgent::GetIntersectionPoint(int power, int turnDelay)
{
	int				i;
	double			dActionRadius;
	double			dBallDistance;
	TVector			tFutureBall;
	TVector			tPrevious;
	TThisPlayer*	pAgent = m_pPerception->GetThisPlayer();
	int				minDir;	
	TDirection		turnDir;
	double			accessDistance;
	CString			msg;

	// ???
	i = 1;
	tPrevious = m_pPerception->GetBall()->GetPosition();
	dActionRadius = pAgent->GetFutureDistanceActive(0, power, pAgent->GetAbsDir()
													, m_pPerception->GetEffort(), m_pPerception->GetDashPowerRate());
	while (i < MAX_FUTURE_POSITIONS)
	{
		// we calculate the intersection of the future agent position and ball position
		tFutureBall = *(m_pPerception->GetFutureBallPosition(i));
		dBallDistance = pAgent->GetDistanceTo(tFutureBall);

		// test if the ball will be outside
		if (m_pPerception->IsOutside(tFutureBall))
		{
			// we run to the last position that is within the field
			// !!! may be if the two distances are big, do not run anyhow
			return tPrevious;
		}
			
		// get distance the agent can dash
		dActionRadius = pAgent->GetNextFutureDistance();
		
		if (m_pPerception->GetThisPlayer()->GetNum() == GOALIE_ID)
			accessDistance = CATCHABLE_AREA_L;
		else
			accessDistance = m_pPerception->GetPlayerSize() + BALL_SIZE + m_pPerception->GetKickableMargin();
		if (dActionRadius + accessDistance > dBallDistance)
		{
			// we have the position, but if the player has to turn, we step one pos further
			minDir = GetDirectionPreciseness(dBallDistance);
			turnDir = pAgent->GetAngleToBody(tFutureBall);

			if ((i < MAX_FUTURE_POSITIONS - turnDelay) && ((turnDir < -minDir) || (turnDir > minDir)))
			{
				// write a debug message
				CString msg;
				msg.Format("Think will have to turn: ballDist %4.1f, minDir %d, turnDir %d"
					, dBallDistance, minDir, (int) turnDir);
				m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);

				// we have to do a turn, so two steps more into future
				tFutureBall =  *(m_pPerception->GetFutureBallPosition(i + turnDelay));
			}
			return tFutureBall;
		}

		// ball will still be too far this setp
		tPrevious = tFutureBall;
		i++;
	}
	
	return tFutureBall;
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::CreateActions()
Date    : 22.02.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: adds the agents behavior function to the network
-----------------------------------------------------------------------------*/
void TDoingAgent::CreateActions()
{
	m_pBehaviorNetwork->AddAction("action", this, (PCallBackAction) DummyAction);
	m_pBehaviorNetwork->AddAction("aNothing", this, (PCallBackAction) DoANothing);
	m_pBehaviorNetwork->AddAction("awaitGoalkick", this, (PCallBackAction) DoAwaitGoalkick);
	m_pBehaviorNetwork->AddAction("catchBall", this, (PCallBackAction) DoCatchBall);
	m_pBehaviorNetwork->AddAction("checkOffside", this, (PCallBackAction) DoCheckOffside);
	m_pBehaviorNetwork->AddAction("kickAwayBall", this, (PCallBackAction) DoKickAwayBall);
	m_pBehaviorNetwork->AddAction("kickGoal", this, (PCallBackAction) DoKickGoal);
	m_pBehaviorNetwork->AddAction("lookStraight", this, (PCallBackAction) DoLookStraight);
	m_pBehaviorNetwork->AddAction("lookToBall", this, (PCallBackAction) DoLookToBall);
	m_pBehaviorNetwork->AddAction("mindBall", this, (PCallBackAction) DoMindBall);
	m_pBehaviorNetwork->AddAction("moveToHomePosition", this, (PCallBackAction) DoMoveToHomePosition);
	m_pBehaviorNetwork->AddAction("passBall", this, (PCallBackAction) DoPassBall);
	m_pBehaviorNetwork->AddAction("protectBall", this, (PCallBackAction) DoProtectBall);
	m_pBehaviorNetwork->AddAction("pushBackward", this, (PCallBackAction) DoPushBackward);
	m_pBehaviorNetwork->AddAction("pushForward", this, (PCallBackAction) DoPushForward);
	m_pBehaviorNetwork->AddAction("relax", this, (PCallBackAction) DoRelax);
	m_pBehaviorNetwork->AddAction("retreat", this, (PCallBackAction) DoRetreat);
	m_pBehaviorNetwork->AddAction("runForwardToBall", this, (PCallBackAction) DoRunForwardToBall);
	m_pBehaviorNetwork->AddAction("runToBall", this, (PCallBackAction) DoRunToBall);
	m_pBehaviorNetwork->AddAction("runToBallGoalie", this, (PCallBackAction) DoRunToBallGoalie);
	m_pBehaviorNetwork->AddAction("runToHomePosition", this, (PCallBackAction) DoRunToHomePosition);
	m_pBehaviorNetwork->AddAction("runWithBall", this, (PCallBackAction) DoRunWithBall);
	m_pBehaviorNetwork->AddAction("sayPosition", this, (PCallBackAction) DoSayPosition);
	m_pBehaviorNetwork->AddAction("searchBall", this, (PCallBackAction) DoSearchBall);
	m_pBehaviorNetwork->AddAction("searchFreeTeammate", this, (PCallBackAction) DoSearchFreeTeammate);
	m_pBehaviorNetwork->AddAction("stopBall", this, (PCallBackAction) DoStopBall);
	m_pBehaviorNetwork->AddAction("turnAroundBallForKickIn", this, (PCallBackAction) DoTurnAroundBallForKickIn);
	m_pBehaviorNetwork->AddAction("waitForKickOff", this, (PCallBackAction) DoWaitForKickOff);

	// inserted for compatibility with agent3
	m_pBehaviorNetwork->AddAction("runFree", this, (PCallBackAction) DoRunFree);
	m_pBehaviorNetwork->AddAction("runToOwnGoal", this, (PCallBackAction) DoRunToOwnGoal);
}

/*-----------------------------------------------------------------------------
Function: Action Functions (sorted alphabetically !!)
Date    : 16.01.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: following are the action functions, that is the basic cognitively
				 uninterruptable behaviors of the agent
-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
does nothing
-----------------------------------------------------------------------------*/
void TDoingAgent::DoANothing(double activation)
{
	CString msg;
	// this is for statistics, to be able to tell how much of the 
	// competence actions have been performed by the server.
	//	Turn(0);
	msg.Format("doNothing (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;
}

/*-----------------------------------------------------------------------------
behavior for offenders to wait until a goalkick has been executed
-----------------------------------------------------------------------------*/
void TDoingAgent::DoAwaitGoalkick(double activation)
{
	int	x;
	int	y = GetPosHome().y;
	int	power;
	CString msg;

	msg.Format("awaitGoalkick (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	if (GetSide() == LEFT_TO_RIGHT)
		x = 35;
	else
		x = -35;
	SetPosHome(CPoint(x, y));

	// run to (new) home position
	if (m_pPerception->GetStamina() > 2500)
		power = 60;
	else
		power = m_pPerception->GetStaminaIncMax() - 5;
	
	if (ComplexDashTo(TVector(x, y), power, 2, TRUE))
		// agent is already there
		LookToBall();
}

/*-----------------------------------------------------------------------------
called by the goalie to catch the ball with hands
-----------------------------------------------------------------------------*/
void TDoingAgent::DoCatchBall(double activation)
{
	TVector	ballInFuture;
	int		ballDir;
	double	ballDist;
	CString	msg;

	// catch to where the ball will be
	ballInFuture = *m_pPerception->GetFutureBallPosition(0);
	if (m_pPerception->GetBall()->IsValid())
		ballDir = m_pPerception->GetBall()->GetDirection() + m_pPerception->GetHeadAngle();
	else
		ballDir = m_pPerception->GetThisPlayer()->GetAngleToBody(ballInFuture);
	ballDist = m_pPerception->GetThisPlayer()->GetDistanceTo(ballInFuture);
	
	msg.Format("catch (%4.2f): ballDist: %4.2f ballDir: %d"
		, activation, ballDist, ballDir);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;
	
	Catch(ballDir);
}

/*-----------------------------------------------------------------------------
checks, if this agent is in offside
-----------------------------------------------------------------------------*/
void TDoingAgent::DoCheckOffside(double activation)
{
	// player does not know about offside, so look left and right
	int			count = m_pPerception->GetCheckOffsideCounter();
	TDirection	turnDir;
	CString msg;

	msg.Format("checkOffside (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	// increment the counter
	m_pPerception->SetCheckOffsideCounter(++count);

	if (count == 1)
	{
		// we can not be sure to have seen complete part of field, so look
		if (GetSide() == LEFT_TO_RIGHT)
			turnDir = - 45 - m_pPerception->GetThisPlayer()->GetAbsDir();
		else
			turnDir = - 135 - m_pPerception->GetThisPlayer()->GetAbsDir();

		// turnhead
		Turn(turnDir);
		SetWaitForNextSee();
	}
	else if (count == 2)
	{
		// we can not be sure to have seen complete part of field, so look 90
		if (GetSide() == LEFT_TO_RIGHT)
			turnDir = 45 - m_pPerception->GetThisPlayer()->GetAbsDir();
		else
			turnDir = 135 - m_pPerception->GetThisPlayer()->GetAbsDir();

		// turnhead
		Turn(turnDir);
		SetWaitForNextSee();
	}
	else
	{
		// player now should know to be in offside or not
		m_pPerception->SetCheckOffsideCounter(0);
		m_pPerception->KnowOffside();
		DoRetreat(activation);
	}
}

/*-----------------------------------------------------------------------------
kicks away the ball. Depending on the position of the agent, the ball is
kicked into goal direction, along the lines or away from the own goal to the side
-----------------------------------------------------------------------------*/
void TDoingAgent::DoKickAwayBall(double activation)
{
	// test the direction to the left and right
	TThisPlayer*	thisPlayer = m_pPerception->GetThisPlayer();
	int				side = m_pPerception->GetSide();
	int				x = (int) thisPlayer->GetAbsX();
	int				y = (int) thisPlayer->GetAbsY();
	TDirection		direction;
	TDirection		dir;
	double			dist;
	double			optimalDist;
	int				optimalDir;
	int				power = 100;
	CString			msg;

	msg.Format("kickAwayBall (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	if (side == RIGHT_TO_LEFT)
	{
		// switch the coordinates of the player
		x = -x;
		y = -y;
	}

/*	// get optimal kickAway direction
	if (thisPlayer->GetDistanceTo(m_pPerception->GetOtherGoal()) < 20.0)
	{
		DoKickGoal(activation);
		return;
	}
*/
	if (x > 28)
	{
		// we are close to the goal, pass to 11m point
		if ((x > 42) && ((y < -12) || (y > 12)))
			direction = m_pPerception->GetObjectDirection(m_pPerception->GetOther11mPoint());		
		else if ((y < -16) || (y > 16))
			direction = m_pPerception->GetObjectDirection(m_pPerception->GetOther11mPoint());
		else
		{
			DoKickGoal(activation);
			return;
		}
	}
	else if (x > -20)
	{
		if (y < -13)
			direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal());
		else if (y < 0)
			// play along the line
			direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherRightPenaltyCorner());
		else if (y > 13)
			direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal());
		else 
			direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherLeftPenaltyCorner());
		// in case we are close to the midline, we do not do a too powerfull kick since the goalie
		// would get it then
		power = 80 - x;
	}
	else if (x > -35)
	{
		if (y < 0)
			// play along the line
			direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherLeftPenaltyCorner());
		else 
			direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherRightPenaltyCorner());
	} 
	else if (y < -13)
		// play along the line
		direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherLeftPenaltyCorner());
	else if (y < 0)
			direction = thisPlayer->GetRelativeFromAbsolutePlayDirection(-40);
	else if (y > 13)
		// play along the line
		direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherRightPenaltyCorner());
	else
		direction = thisPlayer->GetRelativeFromAbsolutePlayDirection(40);

	// take opponents into account (but only if closer than 30m)
	optimalDir = direction;
	optimalDist = m_pPerception->GetDirectionDistance(optimalDir);

	if (optimalDist < 30.0)
	{
		int iCount = 8;
		dir = direction - (5 * iCount);
		for (int i = 0; i <= iCount; i++)
		{
			dist = m_pPerception->GetDirectionDistance(dir);
			if (dist > optimalDist)
			{
				optimalDir = dir;
				optimalDist = dist;
			}
			dir = dir + 10;
		}
	}

	if (power < 80)
		// we do not kick too hard at the midline
		ComplexKick(power, optimalDir);
	else
		// we kick hard
		HardKick(optimalDir);

	msg.Format("Optimal kickaway dir: %d, optimal dir %d, dist %4.1f"
		, (int)direction, (int)optimalDir, optimalDist);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_MIDACTION);
}

/*-----------------------------------------------------------------------------
kicks the ball into that half of the goal, where less players are
-----------------------------------------------------------------------------*/
void TDoingAgent::DoKickGoal(double activation)
{
	int			dirLeft = m_pPerception->GetObjectDirection(m_pPerception->GetOtherLeftGoalPost());
	int			dirMitte = m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal());
	int			dirRight = m_pPerception->GetObjectDirection(m_pPerception->GetOtherRightGoalPost());
	TDirection	deltaLeft = dirMitte - dirLeft;
	TDirection	deltaRight = dirRight - dirMitte;
	TDirection	dirKick;
	CString		msg;
	TPlayer*	pTeammate = m_pPerception->GetFreeTeammate();

	msg.Format("kickGoal (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;


	// if the player has no chance to score because of the goalie, we try to pass
	if ((m_deltaRight < 0) && (m_deltaLeft < 0) 
		&& (pTeammate != NULL) && (pTeammate->GetFreePassDegree() > 0.1))
	{
		DoPassBall(activation);
		return;
	}

	int			playersLeft = m_pPerception->GetPlayersWithin(dirLeft, dirMitte);
	int			playersRight = m_pPerception->GetPlayersWithin(dirMitte, dirRight);


	if (m_deltaLeft > 3)
		deltaLeft = (int) m_deltaLeft / 2;
	else if (m_deltaLeft > 0)
		deltaLeft = 2;
	else
		deltaLeft = (int) deltaLeft / 4;

	if (m_deltaRight > 3)
		deltaRight = (int) m_deltaRight / 2;
	else if (m_deltaRight > 0)
		deltaRight = 2;
	else
		deltaRight = (int) deltaRight / 4;


	if (playersLeft < playersRight)
		// if there is anyone, kick into left half of goal
		dirKick = dirLeft + deltaLeft;
	else if (playersLeft > playersRight)
		// kick into right half
		dirKick = dirRight - deltaRight;
	else if (playersLeft == 0)
		// no players in way, so middle of the goal
		dirKick = dirMitte;
	else if (deltaLeft > deltaRight)
		// kick to the wider half of the goal
		dirKick = dirLeft + deltaLeft;
	else
		// kick to the wider half of the goal
		dirKick = dirRight - deltaRight;

	// no hardkick if we are close to the goal
	if (m_pPerception->GetObjectDistance(m_pPerception->GetOtherGoal()) > 9.0)
		HardKick(dirKick);
	else
		ComplexKick(100, dirKick);
}

/*-----------------------------------------------------------------------------
turns head into running direction
-----------------------------------------------------------------------------*/
void TDoingAgent::DoLookStraight(double activation)
{
	int				headAngle = m_pPerception->GetHeadAngle();
	CString			msg;

	msg.Format("lookStraight (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);

	if (headAngle != 0)
		TurnNeck(-headAngle);
}

/*-----------------------------------------------------------------------------
this could implement a behavior, where the agent has got enough time to
look around every while
-----------------------------------------------------------------------------*/
void TDoingAgent::DoLookToBall(double activation)
{

	// check if we are on a valid position
	CPoint		newPos;
//	int			x = GetPosHome().x;
//	int			y = GetPosY();
	int			power;
	int			stamina;
	int			number		= m_pPerception->GetThisPlayer()->GetNum();
	CString		msg;

	msg.Format("lookToBall (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

//	if (number <= MAX_MIDFIELDER)
		newPos = CalculateDefendingPosition();
//	else
//		newPos = CPoint(GetPosHome().x, GetPosY());

	// check stamina
	stamina = m_pPerception->GetStamina();
	if (stamina > 3000)
		power = 60;
	else if (stamina > 2000)
		power = m_pPerception->GetStaminaIncMax() - 5;
	else
		power = 10;

	// set the (new) home pos
	SetPosHome(newPos);

	// run to (new) home position
	if (ComplexDashTo(TVector(newPos.x, newPos.y), power, 3, TRUE))
		// agent is already there
		LookToBall();
}

/*-----------------------------------------------------------------------------
this implements a behavior, where the agent has got enough time to
look around every while
-----------------------------------------------------------------------------*/
void TDoingAgent::LookToBall()
{
	CString	msg = "lookToBall";
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_MIDACTION);

	// look around or to the ball
	if (!CheckSourrounding())
		// usually this is not allowed!! mid -> high
		DoSearchBall(1.0);
}

/*-----------------------------------------------------------------------------
keeps the eyes pointed to the ball if possible else into rundirection
-----------------------------------------------------------------------------*/
void TDoingAgent::DoMindBall(double activation)
{
	TDirection	ballDir = m_pPerception->GetObjectDirection(m_pPerception->GetBall());
	int			headAngle = m_pPerception->GetHeadAngle();
	TDirection	deltaDir = ballDir + headAngle;
	CString			msg;

	msg.Format("mindBall (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	if ((deltaDir < -135) || (deltaDir > 135))
		// the head can not be turned
		DoLookStraight(activation);
	else
		TurnNeck(ballDir);
}

/*-----------------------------------------------------------------------------
the goalie can move to its home position after a catch command
-----------------------------------------------------------------------------*/
void TDoingAgent::DoMoveToHomePosition(double activation)
{
	int			ballY = (int) m_pPerception->GetBall()->GetAbsY();
	TDirection	dir;
	CString		msg;
	TVector		pos;

	msg.Format("moveToHomePosition (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	// move to 5m corner
	if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
	{
		if (ballY > 0)
		{
			// left top 5m corner
			pos = TVector(-48, -9);
			dir = m_pPerception->GetObjectDirection(m_pPerception->GetOwnRightPenaltyCorner());
		}	
		else
		{
			// left bottom 5m corner
			pos = TVector(-48, 9);
			dir = m_pPerception->GetObjectDirection(m_pPerception->GetOwnLeftPenaltyCorner());
		}
	}
	else
	{
		if (ballY > 0)
		{
			// right top 5m corner
			pos = TVector(-48, 9);
			dir = m_pPerception->GetObjectDirection(m_pPerception->GetOwnLeftPenaltyCorner());
		}
		else
		{
			// right bottom 5m corner
			pos = TVector(-48, -9);
			dir = m_pPerception->GetObjectDirection(m_pPerception->GetOwnRightPenaltyCorner());
		}
	}
	
	GoalieMove(pos, dir);

	// wait a few cycles
	m_iMayWaitTime = m_pPerception->GetSenseTime() + 50;

	// and prevent the agent from catching again
	m_pPerception->SetLastCatchTime(m_pPerception->GetSenseTime() + 10);
}


/*-----------------------------------------------------------------------------
kicks the ball direct to the free teammate
-----------------------------------------------------------------------------*/
void TDoingAgent::DoPassBall(double activation)
{
	TPlayer*	pTeammate = m_pPerception->GetFreeTeammate();
	CString		msg;

	msg.Format("passBall (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	if (pTeammate != NULL)
	{
		// we can pass the ball
		double		distance = m_pPerception->GetObjectDistance(pTeammate);
		int			power = GetPassPower(distance);
		TDirection	dir = pTeammate->GetFreePassDirection();
		TDirection	goalDir = m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal());
		int			passArea = pTeammate->GetFreePassArea();
		int			playerX = (int) m_pPerception->GetThisPlayer()->GetAbsX();
		int			playerY = (int) m_pPerception->GetThisPlayer()->GetAbsY();
		
		if (GetSide() == RIGHT_TO_LEFT)
		{
			playerX = -playerX;
			playerY = -playerY;
		}

		// if we are in opponents half and pass from the side, we pass a little in front of the player
		if ((passArea > 20) && (distance > 12.0) && (playerX > 0) && ((playerY < -10) || (playerY > 10)))
		{
			// we try to play the ball into the run direction of the teammate
			if (goalDir.IsLeftOf(dir))
			{
				// we pass a little more to the left
				dir = dir - (passArea / 3);
			}
			else
			{
				// we pass a little more to the rightt
				dir = dir + (passArea / 3);
			}
		}

/*		
		if (power < 100)
			ComplexKick(power, dir);
		else
			HardKick(dir);
*/
		// look into pass direction before passing
		int			headAngle = m_pPerception->GetHeadAngle();
		TDirection	dirToBody = dir + headAngle;
		double		ballSpeed = m_pPerception->GetBall()->GetAbsSpeed().GetLength();
		double		ballDist  = m_pPerception->GetObjectDistance(m_pPerception->GetBall());

		if ((ballSpeed < 0.1) && (ballDist < 0.7) && ((dirToBody < -20) || (dirToBody > 20)))
		{
			// we have time to look into pass direction
			Turn(dir);
			TurnNeck(-headAngle);
			msg.Format("Pass ball look %d dir: %d free:%4.2f", pTeammate->GetNum(),
				(int) dir, pTeammate->GetFreePassDegree());
			m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		}
		else
		{
			// now pass
			if (power < 100)
				ComplexKick(power, dir);
			else
				HardKick(dir);

			msg.Format("Passed ball to player %d dir: %d free:%4.2f ballSpeed: %4.2f", pTeammate->GetNum(),
				(int) dir, pTeammate->GetFreePassDegree(), ballSpeed);
			m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		}
	}
	else
		// there is no free teammate (should not happen)
		DoKickAwayBall(activation);
}


/*-----------------------------------------------------------------------------
keeps the ball away from an opponent
-----------------------------------------------------------------------------*/
void TDoingAgent::DoProtectBall(double activation)
{
	// !!! test
	DoRunWithBall(activation);
	return;
	
	TPlayer*			pOpp = m_pPerception->GetOpponentAtMe();
	TThisPlayer*	pAgent = m_pPerception->GetThisPlayer();
	int				iOrientation = 0;
	CString			msg;

	m_pView->DisplayDebug("protectBall", KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	// only called if opponent near
	//if ((pOpp != NULL) && (pAgent->GetDistanceTo(pOpp) < SECURITY_DISTANCE))
	if (pOpp != NULL)
	{
		TPlayer*	pOpp2 = m_pPerception->Get2OpponentAtMe();
		int		direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal());
		TPlayer*	pPlayerInWay = m_pPerception->GetPlayerInWay(direction, 5.0);

		if (pOpp2 != NULL)
		{
			// there are 2 opponents around me
			msg.Format("Attacked by 2 Opponents: %d and %d", pOpp->GetNum(), pOpp2->GetNum());
			m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
			// if one is in my way, I kick away the ball
			if ((pPlayerInWay == pOpp2) || (pPlayerInWay == pOpp))
			{
				// when facing an opponent, we kick away the ball or pass it to other agent
				TPlayer* pPlayer = m_pPerception->GetFreeTeammate();
				if ((pPlayer != NULL) && (pPlayer->GetFreePassDegree() > 0.1))
					DoPassBall(activation);
				else
					DoKickAwayBall(activation);
			}
			else
				DoRunWithBall(activation);
		}
		else
		{
			// there is only one opponent, so keep ball
			TDirection	oppDir = pAgent->GetAngleTo(pOpp);
			TDirection	deltaDir = pAgent->GetAngleTo(m_pPerception->GetBall()) - oppDir;
			msg.Format("Attacked by 1 Opponent: %d ", pOpp->GetNum());
			m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);

			if (deltaDir > 0)
				iOrientation = CLOCKWISE;
			else
				iOrientation = COUNTERCLOCKWISE;

			if (!MoveBallToPoint(TDirection(oppDir + 180), iOrientation))
			{
				// the ball is already save
				// !!! perhaps should do something
			}
		}
	}
	else
		// should not happen
		DoRunWithBall(activation);
}

/*-----------------------------------------------------------------------------
defenders call this to move the offside trap backward
-----------------------------------------------------------------------------*/
void TDoingAgent::DoPushBackward(double activation)
{
	CPoint newPos = CalculateDefendingPosition();
	TPlayer* pOpp = m_pPerception->GetOpponentToMark();
/*
	// going back by defenders
	int	x = GetPosHome().x;
	int	y = GetPosHome().y;
	TPlayer*	teammate;
	int	oppX;
	int	xOffside;
	int	xBall = (int) m_pPerception->GetBall()->GetAbsX();;
	int	x2 = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int	number = m_pPerception->GetThisPlayer()->GetNum();
	int	border;
	int	power;
	CString msg;

	msg.Format("pushBackward (%4.2f)", activation);
	m_pView->DisplayDebug(msg , KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	// get offside line
	teammate = m_pPerception->Get2BackmostTeammate();
	if (teammate != NULL)
		xOffside = (int) teammate->GetAbsX();
	else 
		xOffside = xBall;

	// make coordinates side safe
	if (GetSide() == RIGHT_TO_LEFT)
	{
		xOffside = -xOffside;
		xBall = -xBall;
		x2 = -x2;
	}

	if (xBall < xOffside)
		// if the ball is closer to the own goal than our defenders, push backwards
		xOffside = xBall - 3;

	// but do not move offside line to far back
	if (xOffside < -47)
		xOffside = -47;

	if (pOpp != NULL)
	{
		// we have to mark an opponent
		oppX = (int) pOpp->GetAbsX();
		if (GetSide() == RIGHT_TO_LEFT)
			oppX = -oppX;

		// if the ball is very near own goal and playmode not playon we move behind opp
		if ((xBall < -35) && (m_pPerception->GetPlayMode() != PLAY_ON))
			xOffside = oppX - 5;
		else if (oppX > xOffside)
			// only mark if it is not in offside
			xOffside = oppX - 3;

		int oppY = (int) pOpp->GetAbsY();
		if (GetSide() == RIGHT_TO_LEFT)
			// opponent is offside, so we just adapt y coordinate to him
			newPos = CPoint(-xOffside, oppY);
		else
			// opponent is offside, so we just adapt y coordinate to him
			newPos = CPoint(xOffside, oppY);

		if (m_pPerception->GetStamina() > 2000)
			power = 100;
		else if (m_pPerception->GetStamina() > 1500)
			power = 60;
		else 
			power = m_pPerception->GetStaminaIncMax() - 5;

		msg.Format("Mark Opponent %d at (%d, %d)", pOpp->GetNum(), newPos.x, newPos.y);
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_MIDACTION);
	}
	else
	{
		// there is no opponent to hide
		if (number <= MAX_DEFENDER)
		{
			// defender
			border = -40;
			if ((border < xBall) && 
				((m_pPerception->GetPlaymodeOffsideOther()) ||
				(m_pPerception->GetPlaymodeFreekickOwn())))
			{
				border = xBall;
				msg.Format("Offside of other team: move back to %d", xBall);
				m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_MIDACTION);
			}
		}
		else
		{
			// midfielder keeps 10m before defender or ball
			border = xOffside + 10;
			if (border < -25)
				// midfielder maximally push back to x = -20
				border = -25;
		}
		
		// we move to normal y coordinate
		if (GetSide() == RIGHT_TO_LEFT)
			newPos = CPoint(-border, GetPosY());
		else
			newPos = CPoint(border, GetPosY());

		if (m_pPerception->GetStamina() > 3500)
			power = 100;
		else if (m_pPerception->GetStamina() > 2500)
			power = 60;
		else if (m_pPerception->GetStamina() > 1500)
			power = m_pPerception->GetStaminaIncMax() - 5;
		else
			power = 10;

		msg.Format("Push back to (%d, %d)", newPos.x, newPos.y);
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_MIDACTION);

	}
*/
	CString msg;

	msg.Format("pushBackward (%4.2f)", activation);
	m_pView->DisplayDebug(msg , KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	int stamina = m_pPerception->GetStamina();
	int	power;
	if (pOpp != NULL)
	{
		if (stamina > 2000)
			power = 100;
		else if (stamina > 1500)
			power = 60;
		else 
			power = m_pPerception->GetStaminaIncMax() - 5;
	}
	else
	{
		if (stamina > 3500)
			power = 100;
		else if (stamina > 2500)
			power = 60;
		else if (stamina > 1500)
			power = m_pPerception->GetStaminaIncMax() - 5;
		else
			power = 10;
	}

	// set goal position
	SetPosHome(newPos);

	// run to (new) home position
	if (ComplexDashTo(TVector(newPos.x, newPos.y), power, 1, TRUE))
		// agent is already there
		LookToBall();
}

/*-----------------------------------------------------------------------------
called to build up an offside trap by defenders
-----------------------------------------------------------------------------*/
void TDoingAgent::DoPushForward(double activation)
{
	int			power;
	CPoint		newPos		= CalculateDefendingPosition();
	TPlayer*	teammate	= m_pPerception->Get2BackmostTeammate();
	CString		msg;

	msg.Format("pushForward (%4.2f) x:%d y:%d", activation, newPos.x, newPos.y);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	if ((teammate != NULL) && !teammate->IsThisAgent())
		power = PushForwardPower(teammate->GetStamina());
	else
		power = PushForwardPower(m_pPerception->GetStamina());

	// run to (new) home position
	if (ComplexDashTo(TVector(newPos.x, newPos.y), power, 2, TRUE))
		// agent is already there
		LookToBall();
}

int TDoingAgent::PushForwardPower(int stamina)
{
	int power;
	if (stamina > 3000)
		power = 100;
	else if (stamina > 2500)
		power = 70;
	else if (stamina > 2000)
		power = m_pPerception->GetStaminaIncMax() - 5;
	else
		power = 10;
	return power;
}

/*-----------------------------------------------------------------------------
does nothing to get more stamina
-----------------------------------------------------------------------------*/
void TDoingAgent::DoRelax(double activation)
{
	CString msg;

	msg.Format("relax (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;
	LookToBall();
}

/*-----------------------------------------------------------------------------
an offensive player goes backward to leave offside
-----------------------------------------------------------------------------*/
void TDoingAgent::DoRetreat(double activation)
{
	// player is in offside, so go back
	CPoint		newPos		= CalculateDefendingPosition();
	int			xMe			= (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int			deltaX		= newPos.x - xMe;
	int			stamina;
	int			power;
	CString		msg;

	msg.Format("retreat (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	if (deltaX < 0)
		deltaX = -deltaX;

	stamina = m_pPerception->GetStamina();
	if (stamina > 3000)
		power = 100;
	else if (stamina < 1500)
		power = 0;
	else
	{
		power = stamina / 20 - deltaX *	5;
		if (power > 100)
			power = 100;
		else if (power < 10)
			power = 0;
	}

/*	int			x = GetPosHome().x;
	int			y = GetPosY();
	TPlayer*	opponent = m_pPerception->Get2BackmostOpponent();

	// check stamina
	stamina = m_pPerception->GetStamina();
	if (opponent != NULL)
	{
		int	xMe = (int) m_pPerception->GetThisPlayer()->GetAbsX();
		int	x1 = (int) opponent->GetAbsX();
		if (GetSide() == RIGHT_TO_LEFT)
		{
			x1 = -x1;
			xMe = -xMe;
		}

		// get offside line
		int	deltaX = x1 - xMe;
		if (deltaX < 0)
			deltaX = -deltaX;

		if (stamina > 3000)
			power = 100;
		else if (stamina < 1500)
			power = 0;
		else
		{
			power = stamina / 20 - deltaX *	5;
			if (power > 100)
				power = 100;
			else if (power < 10)
				power = 0;
		}
	}
	else
	{
		if (stamina > 2000)
			power = 100;
		else if (stamina > 1500)
			power = m_pPerception->GetStaminaIncMax() - 5;
		else
			power = 0;
	}
*/
	if (power == 0)
	{
		// we do not retreat if we are tired
		LookToBall();
		return;
	}
/*
	if (GetSide() == LEFT_TO_RIGHT)
	{
		if (x > -3)
			// agent is on other side
			SetPosHome(CPoint((x -= 5), y));
	}
	else if (x < 3)
			SetPosHome(CPoint((x += 5), y));
*/
	// run to (new) home position
	SetPosHome(newPos);
	if (ComplexDashTo(TVector(newPos.x, newPos.y), power, 1, TRUE))
		// agent is already there
		LookToBall();
}

/*-----------------------------------------------------------------------------
behavior for offenders to keep on the height of the ball
-----------------------------------------------------------------------------*/
void TDoingAgent::DoRunForwardToBall(double activation)
{
/*
	int	x = GetPosHome().x;
	int	y = GetPosY();
	int	x1 = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int	x2 = (int) m_pPerception->GetBall()->GetAbsX();
	int	power;
	CString msg;
	int	border = 41;

	msg.Format("runForwardToBall (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	if (m_pPerception->GetPlaymodeFreekickOther())
		// if the others have a freekick we do not move forward
		border = 33;

	if (GetSide() == LEFT_TO_RIGHT)
	{
		if ((x2 > x1 + 5) && (x1 < border))
			// agent is behind ball
			x = x1 + 5;
	}
	else if ((x2 < x1 - 5) && (x1 > -border))
			x = x1 - 5;
*/

	CPoint		newPos;
	int			power;
	CString		msg;

	msg.Format("runForwardToBall (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	newPos = CalculateDefendingPosition();

	// run to (new) home position
	if (m_pPerception->GetStamina() > 2000)
		power = 100;
	else
		power = m_pPerception->GetStaminaIncMax() - 5;

	SetPosHome(newPos);
	if (ComplexDashTo(TVector(newPos.x, newPos.y), power, 2, TRUE))
		// agent is already there
		LookToBall();
}

/*-----------------------------------------------------------------------------
player runs to the expected position of the ball
-----------------------------------------------------------------------------*/
void TDoingAgent::DoRunToBall(double activation)
{
	TVector			tFutureBall;
	TVector			agentNextTime;
	CString			msg;
	int				power;
	double			futureDistancteToBall;
	double			maxKickDistance = m_pPerception->GetPlayerSize() + BALL_SIZE + m_pPerception->GetKickableMargin();	

	msg.Format("runToBall (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;
	
	// calculate the power
//	power = MIN_DASH + (int) (activation * (100 - MIN_DASH));
	power = 100;
	
	// dash less if freekick
	if ((m_pPerception->GetPlayMode() != PLAY_ON) && !GetOrdinaryGame())
	{
		if (m_pPerception->GetStamina() < LOW_STAMINA)
		{
			// we have too low stamina, so do not dash at all
			LookToBall();
			return;
		}
		else
			// dash slow, it is a freekick
			power = 15;
	}

	// calculate the intersection point
	tFutureBall = GetIntersectionPoint(power);

	// check if ball will be kickable if we do not dash
	agentNextTime = m_pPerception->GetThisPlayer()->GetPosition() +
						m_pPerception->GetThisPlayer()->GetAbsSpeed();
	futureDistancteToBall = agentNextTime.GetDistanceTo(tFutureBall);
	if (futureDistancteToBall < maxKickDistance - 0.1)
	{
		// we will be at the ball without dash
		msg.Format("runToBall, will be there: future distance: %4.2f", futureDistancteToBall);
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);

		DoTurnToBall(activation);
	}
	else
	{
		int	iAbsX = (int) m_pPerception->GetThisPlayer()->GetAbsX();
		int	iAbsY = (int) m_pPerception->GetThisPlayer()->GetAbsY();

		// when running to the ball, we need to see something
		if (m_pPerception->GetObjectDistance(m_pPerception->GetBall()) < 6.0)
		{
			if (m_pPerception->GetViewAngle() == NORMAL_VIEW)
				ChangeView(NARROW_VIEW, 1);
			m_bWantNarrowView = TRUE;
		}

		// special treatment for goalie
/*		if (m_pPerception->GetThisPlayer()->GetNum() == GOALIE_ID)
			tFutureBall = GoalieRunToBall(tFutureBall);
*/
		// special treatment for defenders when behind opponent
		if (m_pPerception->GetSide() == RIGHT_TO_LEFT)
		{
			iAbsX = -iAbsX;
			iAbsY = -iAbsY;
		}
		if (iAbsX < 0)
			tFutureBall = DefenderRunToBall(tFutureBall, iAbsX, iAbsY);

		// execute the dash command
		if (ComplexDashTo(tFutureBall, power, 0, TRUE, FALSE))
		{
			// agent is already there
	//		DoRunWithBall(activation);
			msg.Format("runToBall, are there");
			m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);

			DoTurnToBall(activation);
		}
	}
}

/*-----------------------------------------------------------------------------
goalie runs differently to the ball than other players
-----------------------------------------------------------------------------*/
void TDoingAgent::DoRunToBallGoalie(double activation)
{
	TVector		tFutureBall;
	TVector		tFutureBallDirect;
	int			power;
	TPlayer*	opponentAtBall = m_pPerception->GetOpponentAtBall();
	TPlayer*	opponentAtMe = m_pPerception->GetOpponentAtMe();
	double		closerDistance;
	CString		msg;

	msg.Format("runToBallGoalie (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;
	
	// calculate the intersection point, with one turn delay
	power = 100;
	tFutureBall = GetIntersectionPoint(power, 1);

	// check if this will be inside the penalty area
	double x = tFutureBall.GetX();
	double y = tFutureBall.GetY();

	if (GetSide() == RIGHT_TO_LEFT)
	{
		x = -x;
		y = -y;
	}

	closerDistance = 2.0;
	if ((x > -36.0) || (y < -16.0) || (y > 16.0))
	{
		// we will reach the ball outside the penalty area.
		// this is only allowed, if the goalie will be definitly at the ball before the agent
		closerDistance = 4.0;
	}

	double		distOpponentBall = 100.0;
	double		distOpponentMe = 100.0;
	double		distMeBall = m_pPerception->GetThisPlayer()->GetDistanceTo(tFutureBall);
	double		distBallGoal = m_pPerception->GetOwnGoal()->GetDistanceTo(tFutureBall);
	double		distMeGoal = m_pPerception->GetThisPlayer()->GetDistanceTo(m_pPerception->GetOwnGoal());

	// look for the threatening player
	if (opponentAtBall != NULL)
	{
		distOpponentBall = opponentAtBall->GetDistanceTo(tFutureBall);
	}
	if (opponentAtMe != NULL)
	{
		distOpponentMe = opponentAtMe->GetDistanceTo(tFutureBall);
	}

	if (distOpponentMe < distOpponentBall)
		distOpponentBall = distOpponentMe;

	// look if I am really closer to the ball
	// and only up to 24 m away from goal
	if ((distMeBall > distOpponentBall - closerDistance) || (distBallGoal > 25.0))
	{
		// no, so keep cool
		AdjustGoaliePosition(tFutureBall);
		return;
	}

	// else run to ball, but check if we can run without turning
	// calculate the intersection point, with no turn delay
	tFutureBallDirect = GetIntersectionPoint(power, 0);
	double		ballDist= m_pPerception->GetThisPlayer()->GetDistanceTo(tFutureBallDirect);
	int			minDir	= GetDirectionPreciseness(ballDist);
	int			dir		= m_pPerception->GetThisPlayer()->GetAngleToBody(tFutureBallDirect);
	if ((dir >= -minDir) && (dir <= minDir))
		tFutureBall = tFutureBallDirect;

/*	if (distBallGoal > distMeGoal - 2.0)
	{
		// just dash without turning
		AdjustGoaliePosition(tFutureBallDirect);
		return;
	}
*/
	// else normal run to ball
	if (m_pPerception->GetPlayMode() != PLAY_ON)
	{
		// dash less if freekick
		if (m_pPerception->GetStamina() < LOW_STAMINA)
		{
			// we have too low stamina, so do not dash at all
			LookToBall();
			return;
		}
		else
			power = 15;
	}

	if (WillHaveBallNextStep())
	{
		// we will be at the ball without dash
		msg.Format("runToBall, will be there: future distance: %4.2f", distMeBall);
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		DoTurnToBall(activation);
	}
	else 
	{
		msg.Format("runToBall x:%4.2f y:%4.2f", tFutureBall.GetX(), tFutureBall.GetY());
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);

		if (ComplexDashTo(tFutureBall, power, 0, TRUE, FALSE))
		{
			// execute the dash command
			DoTurnToBall(activation);
			msg.Format("runToBall, are there");
			m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		}
	}
}


/*-----------------------------------------------------------------------------
runs to the current home position of the player. This position may be specified
by other behaviors
-----------------------------------------------------------------------------*/
void TDoingAgent::DoRunToHomePosition(double activation)
{
	TVector tFutureBall = *m_pPerception->GetFutureBallPosition(15);
	CString		msg;

	msg.Format("runToHomePosition (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	AdjustGoaliePosition(tFutureBall);
}

/*-----------------------------------------------------------------------------
runs to the current home position of the player. This position may be specified
by other behaviors
-----------------------------------------------------------------------------*/
void TDoingAgent::OldDoRunToHomePosition(double activation)
{
	int			power;
	int			precision = 2;
	TDirection	dir;
	TVector		posHome;
	CString		msg;

	msg.Format("runToHomePosition (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	if (m_pPerception->GetThisPlayer()->GetNum() == GOALIE_ID)
	{
		// goalie dashes back faster
		power = 100;
		precision = 2;
		
		// and checks, whether to move left or right
		dir = m_pPerception->GetObjectDirection(m_pPerception->GetBall())
				- m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal());
		
		// check, if the goalie should be in a different position
		switch (m_iIndexPosHome)
		{
		case 0:
				if (dir < 45)
					m_iIndexPosHome = 1;
			break;
		case 1:
				if (dir > 55)
					m_iIndexPosHome = 0;
				else if (dir < 20)
					m_iIndexPosHome = 2;
			break;
		case 2:
				if (dir > 25)
					m_iIndexPosHome = 1;
				else if (dir < 5)
					m_iIndexPosHome = 3;
			break;
		case 4:
				if (dir < -25)
					m_iIndexPosHome = 5;
				else if (dir > -5)
					m_iIndexPosHome = 3;
			break;
		case 5:
				if (dir < -55)
					m_iIndexPosHome = 6;
				else if (dir > -20)
					m_iIndexPosHome = 4;
			break;
		case 6:
				if (dir > -45)
					m_iIndexPosHome = 5;
			break;
		default:
				if (dir > 10)
					m_iIndexPosHome = 2;
				else if (dir < -10)
					m_iIndexPosHome = 4;
			break;
		}

		switch(m_iIndexPosHome)
		{
		case 0:
			if (GetSide() == LEFT_TO_RIGHT)
				SetPosHome(CPoint(-51, 8));
			else
				SetPosHome(CPoint(51, -8));
			break;
		case 1:
			if (GetSide() == LEFT_TO_RIGHT)
				SetPosHome(CPoint(-47, 4));
			else
				SetPosHome(CPoint(47, -4));
			break;
		case 2:
			if (GetSide() == LEFT_TO_RIGHT)
				SetPosHome(CPoint(-47, 2));
			else
				SetPosHome(CPoint(47, -2));
			break;
		case 4:
			if (GetSide() == LEFT_TO_RIGHT)
				SetPosHome(CPoint(-47, -2));
			else
				SetPosHome(CPoint(47, 2));
			break;
		case 5:
			if (GetSide() == LEFT_TO_RIGHT)
				SetPosHome(CPoint(-47, -4));
			else
				SetPosHome(CPoint(47, 4));
			break;
		case 6:
			if (GetSide() == LEFT_TO_RIGHT)
				SetPosHome(CPoint(-51, -8));
			else
				SetPosHome(CPoint(51, 8));
			break;
		default:
			if (GetSide() == LEFT_TO_RIGHT)
				SetPosHome(CPoint(-47, 0));
			else
				SetPosHome(CPoint(47, 0));
			break;
		}
	}
	else
	{
		// it is not the goalie, so power depends on stamina
		if (m_pPerception->GetStamina() > 2000)
			power = 30;
		else
			power = 10;
	}

	posHome = TVector(GetPosHome().x, GetPosHome().y);

	if (ComplexDashTo(posHome, power, precision, TRUE))
	{
		// agent is already there
		DoSearchBall(activation);
	}
}

/*-----------------------------------------------------------------------------
the player does a dribbling
-----------------------------------------------------------------------------*/
void TDoingAgent::DoRunWithBall(double activation)
{
	CString	msg;
	int		iAbsX = (int) m_pPerception->GetThisPlayer()->GetAbsX();
	int		iAbsY = (int) m_pPerception->GetThisPlayer()->GetAbsY();
	int		direction;
	TPlayer*	opponent = m_pPerception->GetOpponentAtMe();

	msg.Format("runWithBall (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

// !!! test
//	DribbleCounterclockwise(HARD_DRIBBLE);
//	HardKick(m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal()));
//	return;

	if (m_pPerception->GetSide() != LEFT_TO_RIGHT)
	{
		iAbsX = -iAbsX;
		iAbsY = -iAbsY;
	}

	if (iAbsX > 40)
	{
//		direction = m_pPerception->GetObjectDirection(m_pPerception->GetOther11mPoint());
		DoKickAwayBall(activation);
		return;
	}
	else
	{
		if ((opponent != NULL) && (m_pPerception->GetObjectDistance(opponent) > 7.0))
			direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal());
//		direction = m_pPerception->GetObjectDirection(m_pPerception->GetOtherGoal());

		else if (iAbsY > 28)
			direction = m_pPerception->GetThisPlayer()->GetRelativeFromAbsolutePlayDirection(-20);
		else if (iAbsY < -28)
			direction = m_pPerception->GetThisPlayer()->GetRelativeFromAbsolutePlayDirection(20);
		else
			// in new version we run straight
			direction = m_pPerception->GetThisPlayer()->GetRelativeFromAbsolutePlayDirection(0);
	}

	msg.Format("runwithball direction: %d", direction);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_MIDACTION);
	
	int			headAngle = m_pPerception->GetHeadAngle();
	TDirection	dirToBody = direction + headAngle;
	double		ballSpeed = m_pPerception->GetBall()->GetAbsSpeed().GetLength();

	// check if accessible area is ok
	double	dist = m_pPerception->GetAccessDistance(direction);
	if (dist < 5.0)
	{
		// there is an opponent near or we are tired
		double	dMaxKickDistance = m_pPerception->GetPlayerSize() + BALL_SIZE + m_pPerception->GetKickableMargin();
		double	ballDist = m_pPerception->GetObjectDistance(m_pPerception->GetBall());

		if (ballDist < dMaxKickDistance)
		{
			// ball is kickable
			TPlayer* pPlayer = m_pPerception->GetFreeTeammate();
			if ((pPlayer != NULL) && (pPlayer->GetFreePassDegree() > 0.3))
				DoPassBall(activation);
//				else if (m_pPerception->GetThisPlayer()->IsInOwnHalfOfField())
//			else if (iAbsX < 25)
			else if (iAbsX < 10)
				DoKickAwayBall(activation);
			else
			{
				// check if an opponent is close to the ball
				if (m_pPerception->GetOpponentNearBall(4.0)) 
				{
					DoKickAwayBall(activation);

					msg.Format("run with ball, opponent at ball");
					m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
				}
//				else if (m_pPerception->GetOpponentNearBall(6.0)) 
//					SurroundOpponent(direction);
				else
					// there is no opponent really near, so dribble
					ComplexDashAndKick(100, dist, direction);		
			}
		}
		else
			// the ball is not kickable, so run to it
			ComplexDashAndKick(100, dist, direction);		
	}
	else
	{
		if ((ballSpeed < 0.1) && ((dirToBody < -90) || (dirToBody > 90)))
		{
			// we have time to look into run direction
			Turn(direction);
			TurnNeck(-headAngle);
			msg.Format("RunWithBall first turn dir:%d", direction);
			m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
			return;
		}

		ComplexDashAndKick(100, dist, direction);
	}
}

/*-----------------------------------------------------------------------------
say the own position and that of the valid ball and all valid players
-----------------------------------------------------------------------------*/
void TDoingAgent::DoSayPosition(double activation)
{
	TThisPlayer *pThisPlayer = m_pPerception->GetThisPlayer();
	CString	sMsg;
	CString	sStr1;
	CString	sStr2;
	char		cSep = ' ';
	int		i;

	if (m_bSayDone)
		return;
	m_bSayDone = TRUE;
	m_pView->DisplayDebug("sayPosition", KATEGORY_ACTION, DEBUG_HIGHACTION);
	
	sMsg = GetSideIdentifier();
	sMsg += cSep;

	//recipient of the message
	i = 0;	//all players
	sStr1.Format("%d",i);
	sMsg += sStr1;
	sMsg += cSep;

	//dispatcher of the message
	i = pThisPlayer->GetNum();
	sStr1.Format("%d",i);
	sMsg += sStr1;
	sMsg += cSep;

	//dispatchers stamina
	i = pThisPlayer->GetStamina();
	sStr1.Format("%d",i);
	sMsg += sStr1;
	sMsg += cSep;

	//msg type p for position
	sMsg += "P ";

	//string of valid ball and players
	sStr1 = m_pPerception->GetValidSeen();
	sMsg += sStr1;

	Say(sMsg);
}

/*-----------------------------------------------------------------------------
searches the ball by looking around and changing view angle to 180
-----------------------------------------------------------------------------*/
void TDoingAgent::DoSearchBall(double activation)
{
	TBall*		pBall = m_pPerception->GetBall();
	TDirection	headAngle = m_pPerception->GetHeadAngle();
	CString		msg;

	msg.Format("searchBall (%4.2f)", activation);
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	// turn head into dashing direction
	TurnNeck(-(int)headAngle);

	if (pBall->IsInMind())
	{
		// it is allready seen or at least in mind so turn to the direction
		DoTurnToBall(activation);
	}
	else 
	{
		// we have no idea, where the ball should be
		// turnhead
		Turn(85);
		SetWaitForNextSee();
	}
}

/*-----------------------------------------------------------------------------
searches for a free teammate
-----------------------------------------------------------------------------*/
void TDoingAgent::DoSearchFreeTeammate(double activation)
{
	TBall*		pBall = m_pPerception->GetBall();
	TDirection	headAngle = m_pPerception->GetHeadAngle();
	CString		msg;

	msg.Format("searchFreeTeammate (%4.2f)", activation);	
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	// could be cleverer if we look at optimal pass direction first
	// we have no idea, where a free teammate should be
	// turnhead
	Turn(90);
	SetWaitForNextSee();
}

/*-----------------------------------------------------------------------------
stops the ball
-----------------------------------------------------------------------------*/
void TDoingAgent::DoStopBall(double activation)
{
	TBall*		pBall = m_pPerception->GetBall();
	
	m_pView->DisplayDebug("stopBall", KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;
	
	ComplexKick(0, pBall->GetDirection());
}

/*-----------------------------------------------------------------------------
the player runs behind the ball if playmode is KICK_IN
-----------------------------------------------------------------------------*/
void TDoingAgent::DoTurnAroundBallForKickIn(double activation)
{
	m_pView->DisplayDebug("turnAroundBallForKickIn", KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	// test the direction to the left and right
	double MIN_DISTANCE_TO_BALL = 3.0;
	int KICK_IN_DASH_POWER = 30;
	int MIN_DISTANCE_TO_POSITION = 2;
	TBall*			pBall = m_pPerception->GetBall();
	TVector			posBall = pBall->GetPosition();

	if (posBall.GetY() > 0)
	{
		posBall += TVector(0.0, MIN_DISTANCE_TO_BALL);
		if (ComplexDashTo(posBall,	KICK_IN_DASH_POWER, MIN_DISTANCE_TO_POSITION))
			DoRunToBall(activation);
	}
	else
	{
		posBall -= TVector (0.0, MIN_DISTANCE_TO_BALL);
		if (ComplexDashTo(posBall,	KICK_IN_DASH_POWER, MIN_DISTANCE_TO_POSITION))
			DoRunToBall(activation);
	}
}

/*-----------------------------------------------------------------------------
turns body to the known ball direction
-----------------------------------------------------------------------------*/
void TDoingAgent::DoTurnToBall(double activation)
{
	TVector*		pBallPosition = m_pPerception->GetFutureBallPosition(1);
	int			deltaDir = m_pPerception->GetThisPlayer()->GetAngleToBody(*pBallPosition);

	if ((deltaDir > 8) || (deltaDir < -8))
		// turn body
		Turn(deltaDir);
}

/*-----------------------------------------------------------------------------
the goalie calls this to wait a little at kick off
-----------------------------------------------------------------------------*/
void TDoingAgent::DoWaitForKickOff(double activation)
{
	int			ballY = (int) m_pPerception->GetBall()->GetAbsY();
	int			dir;
	CString		msg;

	msg.Format("waitForKickOff (%4.2f)", activation);	
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_HIGHACTION);
	m_bCommandDone = TRUE;

	// may turn to kick away direction
	if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
	{
		if (ballY > 0)
			// left top 5m corner
			dir = m_pPerception->GetObjectDirection(m_pPerception->GetOwnRightPenaltyCorner());
		else
			// left bottom 5m corner
			dir = m_pPerception->GetObjectDirection(m_pPerception->GetOwnLeftPenaltyCorner());
	}
	else
	{
		if (ballY > 0)
			// right top 5m corner
			dir = m_pPerception->GetObjectDirection(m_pPerception->GetOwnLeftPenaltyCorner());
		else
			// right bottom 5m corner
			dir = m_pPerception->GetObjectDirection(m_pPerception->GetOwnRightPenaltyCorner());
	}

	if (m_iMayWaitTime - m_pPerception->GetSenseTime() == 10)
	{
		// we move to kickaway ball
		TVector	pos;
		int		moveDir;
		int		ballY = (int) m_pPerception->GetBall()->GetAbsY();

		// move to corner
		if (m_pPerception->GetSide() == LEFT_TO_RIGHT)
		{
			if (ballY < 0)
			{
				// left top corner
				pos = TVector(-40, -9);
				moveDir = m_pPerception->GetObjectDirection(m_pPerception->GetOtherRightPenaltyCorner());
			}	
			else
			{
				// left bottom corner
				pos = TVector(-40, 9);
				moveDir = m_pPerception->GetObjectDirection(m_pPerception->GetOtherLeftPenaltyCorner());
			}
		}
		else
		{
			if (ballY < 0)
			{
				// right top corner
				pos = TVector(-40, 9);
				moveDir = m_pPerception->GetObjectDirection(m_pPerception->GetOtherLeftPenaltyCorner());
			}
			else
			{
				// right bottom 5m corner
				pos = TVector(-40, -9);
				moveDir = m_pPerception->GetObjectDirection(m_pPerception->GetOtherRightPenaltyCorner());
			}
		}

		GoalieMove(pos, moveDir);
	}
	else
		// then turn into the kickaway direction 
		Turn(dir);

	// and prevent the agent from catching again
	m_pPerception->SetLastCatchTime(m_pPerception->GetSenseTime() + 5);
}


// Inserted for compatibility with Agent3

void TDoingAgent::DoRunFree(double activation)
{
	// run 20 meters from ball
	TBall*		pBall = m_pPerception->GetBall();
	int			x = (int) pBall->GetAbsX();
	int			y = (int) pBall->GetAbsY();

	if (GetSide() == LEFT_TO_RIGHT) 
		x += 20;
	else
		x -= 20;

	if (ComplexDashTo(TVector(x, y), MIN_DASH + (int) (activation * (100 - MIN_DASH)), 4, TRUE))
		// player is already there
		DoSearchBall();
}
void TDoingAgent::DoRunToOwnGoal(double activation)
{
	int	dirGoal = m_pPerception->GetObjectDirection(m_pPerception->GetOwnGoal());
	int	dirBall = m_pPerception->GetObjectDirection(m_pPerception->GetBall());
	int	delta = dirGoal - dirBall;

	if (delta < 0)
		delta = -delta;
	if (delta < 20)
		ComplexDash(100, dirBall);
	else
		ComplexDash(100, dirGoal);
}



TVector TDoingAgent::CalculateKick(double dir,double bx, double by,double mf)
{
    double kickx;
    double kicky;

	CString		msg;

	msg.Format("CalculateKick dir:%4.2f bx:%4.2f by:%4.2f mf:%4.2f", dir, bx, by, mf);	
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);

//    if (dir != 0 && dir != 180) {
        double dirn = dir; 
        /*if (dir >= 90 && dir < 180) {
            dirn = dirn - 90;
        } */
        dirn *= DEG2RAD;
        dirn = tan(dirn);
        double a = 1 + dirn*dirn;
        double b = -2*bx-2*dirn*by;
        double c = bx*bx + by*by - mf*mf;
        double root = b*b - 4*a*c;
    
		if (root < 0)
		{
            root = 0;
        }
        double x1;
        if (dir <= 90 && dir >= -90) {
            x1 = (-b + sqrt(root)) / (2*a);
        } else {
            x1 = (-b - sqrt(root)) / (2*a);
        }

        double y1 = dirn * x1;
        kickx = x1 - bx;
        kicky = y1 - by;

/*    } else {
        double x1 = 0;
        double p = -2 * by;
        double q = by*by + bx*bx -mf*mf;
        double root = (p*p)/4 - q;
        if (root < 0){
            root = 0;
        }
        double y1 = -p/2 + sqrt(root);
        kickx = x1 - bx;
        kicky = y1 - by;
    }
*/	

	msg.Format("Result CalculateKick x:%4.2f y:%4.2f", kickx, kicky);	
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);

	return TVector(kickx, kicky);
}

// tFutureBall: position where the ball will be when we have it
TVector TDoingAgent::GoalieRunToBall(TVector tFutureBall)
{
	double x = tFutureBall.GetX();
	double y = tFutureBall.GetY();

	if (GetSide() == RIGHT_TO_LEFT)
	{
		x = -x;
		y = -y;
	}

	if ((x > -36.0) || (y < -16.0) || (y > 16.0))
	{
		// we will reach the ball outside the penalty area.
		// this is only allowed, if the goalie will be definitly at the ball before the agent
		TPlayer*	opponentAtBall = m_pPerception->GetOpponentAtBall();
		TPlayer*	opponentAtMe = m_pPerception->GetOpponentAtMe();
		double		distOpponentBall = 100.0;
		double		distOpponentMe = 100.0;
		double		distMeBall = m_pPerception->GetThisPlayer()->GetDistanceTo(tFutureBall);
		CString		msg;

		// look for the threatening player
		if (opponentAtBall != NULL)
		{
			distOpponentBall = opponentAtBall->GetDistanceTo(tFutureBall);
		}
		if (opponentAtMe != NULL)
		{
			distOpponentMe = opponentAtMe->GetDistanceTo(tFutureBall);
		}

		if (distOpponentMe < distOpponentBall)
			distOpponentBall = distOpponentMe;

		// look if I am really closer to the ball
		if (distMeBall < distOpponentBall - 2.0)
			// yes, so we try a kick
			return tFutureBall;

		// we should wait somewhere 11m away from the goal
		TVector	temp = tFutureBall - m_pPerception->GetOwnGoal()->GetPosition();
		temp.Normalize();
		temp = m_pPerception->GetOwnGoal()->GetPosition() + 10.0 * temp;

		msg.Format("Goalie Run: x:%4.2f y:%4.2f", temp.GetX(), temp.GetY());	
		m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);

		return temp;
	}

	return tFutureBall;
}

/*-----------------------------------------------------------------------------
the defenders and midfielders should try to get behind the opponent if the opponent has
the ball and is closer to the own goal
-----------------------------------------------------------------------------*/
TVector TDoingAgent::DefenderRunToBall(TVector tFutureBall, int playerX, int playerY)
{
	TPlayer*	opponent = m_pPerception->GetOpponentAtBall();

	if (opponent != NULL)
	{
		TVector		ballPos		= m_pPerception->GetBall()->GetPosition();
		double		ballDist	= m_pPerception->GetObjectDistance(m_pPerception->GetBall());
		int			ballDir		= m_pPerception->GetObjectDirection(m_pPerception->GetBall());
		TDirection	ballBodyDir = m_pPerception->GetThisPlayer()->GetAngleToBody(ballPos);
		double		oppDist		= m_pPerception->GetObjectDistance(opponent);
		double		distBallGoal= m_pPerception->GetOwnGoal()->GetDistanceTo(tFutureBall);
		TDirection	oppDir		= m_pPerception->GetObjectDirection(opponent);
		TDirection	delta		= oppDir - ballDir;
		int			yBall		= (int) ballPos.GetY();
		int			side		= GetSide();
		double		borderDist;
		CString		msg;

		if (distBallGoal < 16.0)
			// close to the ball we do not make any compromise
			return tFutureBall;

		delta.Abs();
		if ((oppDist < ballDist) && (delta < 30))
		{
			// the opponent is closer to the ball
			double	oppGoalDist = opponent->GetDistanceTo(m_pPerception->GetOwnGoal());
			double	myGoalDist = m_pPerception->GetObjectDistance(m_pPerception->GetOwnGoal());

			// avoid oszillation
			if ((ballBodyDir > -10) && (ballBodyDir < 10))
				// we are heading to the ball, so we may keep running to the ball
				borderDist = oppDist / 3.0;
			else
				borderDist = oppDist / 2.0;

			double deltaDist = oppGoalDist - myGoalDist - borderDist;
			if (deltaDist < 0.0)
			{
				// calculate the intersection of ball goal line
				TVector	newPos = m_pPerception->GetOwnGoal()->GetPosition() - tFutureBall;
				newPos.Normalize();
				newPos = tFutureBall - deltaDist * newPos;
				tFutureBall = newPos;

/*				
				
				if (side == RIGHT_TO_LEFT)
				{
					yBall = -yBall;
				}
				// we have to try to get closer to the own goal
				if (((yBall < 0) && (yBall > playerY))
					|| ((yBall > 0) && (yBall < playerY)))
				{
					tFutureBall = m_pPerception->GetOwnLeftGoalPost()->GetPosition();

				}
				else
				{
					// run straight back
					if (side == RIGHT_TO_LEFT)
						tFutureBall = TVector(52, -playerY);
					else
						tFutureBall = TVector(-52, playerY);
				}
/*				else if (yBall > 17)
				{
					if (playerY > 15)
						tFutureBall = m_pPerception->GetOwnRightGoalPost()->GetPosition();
					else
					{
						if (side == RIGHT_TO_LEFT)
							tFutureBall = TVector(52, -playerY);
						else
							tFutureBall = TVector(-52, playerY);
					}
				}
				else if (playerY < - 10)
					tFutureBall = m_pPerception->GetOwnLeftGoalPost()->GetPosition();
				else if (playerY > 10)
					tFutureBall = m_pPerception->GetOwnRightGoalPost()->GetPosition();
				else
					tFutureBall = m_pPerception->GetOwnGoal()->GetPosition();
*/
				msg.Format("Defender Run: dball:%4.2f dopp:%4.2f ogoal:%4.2f ichgoal:%4.2f x:%4.2f y:%4.2f"
					, ballDist, oppDist, oppGoalDist, myGoalDist, tFutureBall.GetX(), tFutureBall.GetY());
				m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
			}
		}
	}
	return tFutureBall;
}

/*-----------------------------------------------------------------------------
the goalie can move to its home position after a catch command
-----------------------------------------------------------------------------*/
void TDoingAgent::GoalieMove(TVector pos, int dir)
{
	// then turn into the kickaway direction (remember that the ball is  now on the other side)
	int headAngle = (int) m_pPerception->GetHeadAngle();
	Move((int) pos.GetX(), (int) pos.GetY());
	TurnNeck(-headAngle);
	Turn(dir + headAngle, TRUE);

	m_pPerception->SetSpeed(0.0);
	// update ball
	m_pPerception->SetGoalieCaught(FALSE);
	m_pPerception->GetBall()->SetPosition(m_pPerception->GetThisPlayer()->GetPosition()
														+ TVector(0.1, 0.0));
	// we do it twice to overwrite also oldposition
	m_pPerception->GetBall()->SetPosition(m_pPerception->GetThisPlayer()->GetPosition()
														+ TVector(0.1, 0.0));
	// we do it three times to overwrite also oldoldposition
	m_pPerception->GetBall()->SetPosition(m_pPerception->GetThisPlayer()->GetPosition()
														+ TVector(0.1, 0.0));
	m_pPerception->GetBall()->SetLastSeePosition();
	m_pPerception->GetBall()->SetAbsSpeed(TVector(0.0, 0.0));

	Turn(0, TRUE);
	Turn(0, TRUE);
	Turn(0, TRUE);
	Turn(0, TRUE);
	Turn(0, TRUE);
	Turn(0, TRUE);
}


/*-----------------------------------------------------------------------------
keeps the goalie in line with ball and goal
-----------------------------------------------------------------------------*/
void TDoingAgent::AdjustGoaliePosition(TVector tFutureBall)
{
	// check if we have to run backward
	TThisPlayer*	pAgent	= m_pPerception->GetThisPlayer();
	int		power			= 100;
	int		turnDir			= pAgent->GetAngleToBody(tFutureBall);
	double	distBallGoal	= m_pPerception->GetOwnGoal()->GetDistanceTo(tFutureBall);
	double	distMeGoal		= pAgent->GetDistanceTo(m_pPerception->GetOwnGoal());
	double	absY			= pAgent->GetAbsY();
	TDirection standDir;
	CString	msg;
	
	if ((turnDir < -90) || (turnDir > 90))
	{
		// check if we have enough time to turn, else move backward
		if (distBallGoal - distMeGoal < 10.0)
			// we have to dash backward
			power = -power;	
	}

	// calculate the intersection point
	TVector	newPos = tFutureBall - m_pPerception->GetOwnGoal()->GetPosition();
	newPos.Normalize();
	newPos = m_pPerception->GetOwnGoal()->GetPosition() + 12.0 * newPos;
	double	newX = newPos.GetX(); 
	double	newY = newPos.GetY(); 

	if (GetSide() == RIGHT_TO_LEFT)
	{
		newX = -newX;
		newY = -newY;
		absY = -absY;
	}

	if (newX < -50.0)
		newX = -50.0;

	if (newY < -10.0)
	{
		// hard on left side
		if (absY < newY)
			standDir = pAgent->GetRelativeFromAbsolutePlayDirection(30);
		else
			standDir = pAgent->GetRelativeFromAbsolutePlayDirection(-150);
	}
	else if (newY < -5.0)
	{
		if (absY < newY)
			standDir = pAgent->GetRelativeFromAbsolutePlayDirection(70);
		else
			standDir = pAgent->GetRelativeFromAbsolutePlayDirection(-110);
	}
	else if (newY < 5.0)
	{
		// middle of goal
		newX = -41.0;

		if (absY < newY)
			standDir = pAgent->GetRelativeFromAbsolutePlayDirection(90);
		else
			standDir = pAgent->GetRelativeFromAbsolutePlayDirection(-90);
	}
	else if (newY < 10.0)
	{
		if (absY < newY)
			standDir = pAgent->GetRelativeFromAbsolutePlayDirection(110);
		else
			standDir = pAgent->GetRelativeFromAbsolutePlayDirection(-70);
	}
	else
	{
		if (absY < newY)
			standDir = pAgent->GetRelativeFromAbsolutePlayDirection(150);
		else
			standDir = pAgent->GetRelativeFromAbsolutePlayDirection(-30);
	}

	if (GetSide() == RIGHT_TO_LEFT)
	{
		newX = -newX;
		newY = -newY;
	}
	
	msg.Format("Goalie Run: x:%4.2f y:%4.2f dir:%d", newX, newY, (int) standDir);	
	m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);

	// when running to the ball, we need to see something
	if (distBallGoal < 32.0)
	{
		if (m_pPerception->GetViewAngle() == NORMAL_VIEW)
			ChangeView(NARROW_VIEW, 1);
		m_bWantNarrowView = TRUE;
	}

	newPos = TVector(newX, newY);

	// execute the dash command
	if (ComplexDashTo(newPos, power, 2, TRUE, FALSE))
	{
		// agent is roughly there, so check if we can improve the coordinates by simply dashing
		int		dirToNewPos = pAgent->GetAngleToBody(newPos); 
		double  distToNewPos = pAgent->GetDistanceTo(newPos);
		double	dashDist = distToNewPos * cos(dirToNewPos * DEG2RAD);
		if (((dashDist < -0.5) || (dashDist > 0.5)) && (distBallGoal < 27.0))
		{
			int power = (int) (dashDist / m_pPerception->GetDashPowerRate());
			if (power < -100)
				power = -100;
			else if (power > 100)
				power = 100;
			Dash(power);
			msg.Format("Goalie Run adjust: dashdist:%4.2f dist:%4.2f dir:%d power%d", 
				dashDist, distToNewPos, (int) dirToNewPos, power);	
			m_pView->DisplayDebug(msg, KATEGORY_ACTION, DEBUG_LOWACTION);
		}
		else
		{
			// agent is already there, so just look to the ball
			TDirection neckDir = pAgent->GetAngleTo(*m_pPerception->GetFutureBallPosition(1));
			if (distBallGoal - distMeGoal > 8.0)
			{
				// ball is far enough to turn into standDir
				TDirection actualTurnDir = standDir + m_pPerception->GetHeadAngle();
				if ((actualTurnDir < -10) || (actualTurnDir > 10))
				{
					TDirection actualNeckDir = neckDir - actualTurnDir;
					Turn(actualTurnDir);
					TurnNeck(actualNeckDir);
				}
				else
					TurnNeck(neckDir);
			}
			else
				// just move the head
				TurnNeck(neckDir);
		}
	}
}

